/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.bndmgt.internal;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.wso2.carbon.bndmgt.BundleManagementAgent;

/**
 * Activator for the Bundle Management Agent. This agent is responsible for starting the bundles
 * according to the custom start level manifest header <code>WSO2-Bundle-StartLevel</code>. This
 * process will happen at the initiation of the carbon based products and also as and when there
 * are new bundles available for the environment.
 *
 * @see org.osgi.framework.BundleActivator
 * @see org.wso2.carbon.bndmgt.BundleManagementAgent
 */
public class BundleManagementActivator implements BundleActivator {

    /** Used to store the Log variable reference **/
    private static final Log log = LogFactory.getLog(BundleManagementActivator.class);

    /** Holds the reference to the BundleManagementAgent inside the activator **/
    private BundleManagementAgent bundleManagementAgent;

    /**
     * Initiates the {@link org.wso2.carbon.bndmgt.BundleManagementAgent} and then initializes
     * the start levels for the bundles which contains the custom start level header and then
     * registers a listener to listen for new bundles entering to the environment.
     * 
     * @param bundleContext information about the bundle and the OSGi environment
     * @throws Exception in case of a failure in the start level initialization
     * @see org.osgi.framework.BundleActivator#start(org.osgi.framework.BundleContext) 
     */
    public void start(BundleContext bundleContext) throws Exception {

        if (log.isDebugEnabled()) {
            log.debug("Sarting the Bundle Management Bundle...");
            log.debug("Registering the Bundle Management Agent");
        }
        // initializes the BundleManagementAgent
        bundleManagementAgent = BundleManagementAgent.getInstance(bundleContext);
        
        // register the bundle management agent as the framework listener to start the carbon 
        // bundles as per their start levels
        bundleContext.addFrameworkListener(bundleManagementAgent);
        
        // then register the agent to listen for the bundles that will be installed
        // after the initiallization
        bundleContext.addBundleListener(bundleManagementAgent);

        if (log.isDebugEnabled()) {
            log.debug("Successfully registered the Bundle Management Agent");
            log.debug("Started the Bundle Management Bundle...");
        }
    }

    /**
     * Removes the bundle listener from being listenning for new bundles to resolve start level
     * 
     * @param bundleContext information about the bundle and the OSGi environment
     * @throws Exception in case of a failure in the start level initialization
     * @see org.osgi.framework.BundleActivator#stop(org.osgi.framework.BundleContext)
     */
    public void stop(BundleContext bundleContext) throws Exception {
        
        if (log.isDebugEnabled()) {
            log.debug("Stopping the Bundle Management Agent...");
            log.debug("Removing the registered Bundle Management Agent");
        }

        // Removes the listener for the bundles to listen for start level
        bundleContext.removeBundleListener(bundleManagementAgent);
        bundleManagementAgent = null;
        
        if (log.isDebugEnabled()) {
            log.debug("Successfully removed the Bundle Management Agent");
            log.debug("Stopped the Bundle Management Agent...");
        }
    }
}
