/*
 * Copyright 2004,2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.bridge;

import org.apache.felix.framework.Felix;
import org.apache.felix.framework.cache.BundleCache;
import org.apache.felix.framework.util.FelixConstants;
import org.apache.felix.framework.util.StringMap;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;

import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;
import java.util.jar.Attributes;
import java.util.jar.JarOutputStream;
import java.util.jar.Manifest;

import static org.wso2.carbon.bridge.FrameworlLauncherConstants.*;

/**
 * TODO: this should contain
 */
public class FelixFrameworkLauncher implements FrameworkLauncher {

    /**
     * The property name used to specify an URL to the system
     * property file.
     */
    private static final String SYSTEM_PROPERTIES_PROP = "felix.system.properties";
    /**
     * The default name used for the system properties file.
     */
    private static final String SYSTEM_PROPERTIES_FILE_VALUE = "system.properties";

    /**
     * The default name used for the configuration properties file.
     */
    private static final String CONFIG_PROPERTIES_FILE_VALUE = "config.properties";

    /**
     *
     */
    private static final String RESOURCE_BASE = "/WEB-INF/felix/";


    /**
     * System bundle
     */
    private Felix felix = null;

    /**
     * Servletcontext
     */
    private ServletContext context;

    /**
     *
     */
    private File platformDirectory;

    /**
     *
     */
    private ServletConfig config;

    public void init(ServletConfig servletConfig) {
        context = servletConfig.getServletContext();
        config = servletConfig;
    }

    public void start() {

        if (platformDirectory == null) {
            throw new IllegalStateException(
                    "Could not start the Framework - (not deployed)");
        }
        
        File plugins = new File(platformDirectory, "plugins");
        // This system property is refereced by config.properties file. 
        System.setProperty("work.plugin.dir",
                           plugins.getAbsolutePath());
        String orginalLocation = RESOURCE_BASE + "plugins/";
        copyResource(orginalLocation, plugins);
        //copying rest of the plugins to the work area.
        copyResource("/WEB-INF/plugins/", plugins);
        // Load system properties.
        loadSystemProperties();
        try {
            // Read configuration properties.
            Properties configProps = loadConfigProperties();

            // Copy framework properties from the system properties.
            copySystemProperties(configProps);

            // Print welcome banner.
            System.out.println("\nWelcome to Felix ...");

            // Create a list for custom framework activators and
            // add an instance of the auto-activator it for processing
            // auto-install and auto-start properties.
            List list = new ArrayList();
            list.add(new AutoActivator(configProps));
            // Create a case-insensitive property map.
            Map configMap = new StringMap(configProps, false);
            configMap.put(FelixConstants.EMBEDDED_EXECUTION_PROP, "true");
            configMap.put(BundleCache.CACHE_PROFILE_DIR_PROP, plugins.getAbsolutePath() + "/cache");
            // Create an instance of the framework.
            ClassLoader prevCls = Thread.currentThread().getContextClassLoader();
            try {
                ClassLoader newCls = new ContextFinder(prevCls);
                Thread.currentThread().setContextClassLoader(newCls);
                felix = new Felix(configMap, list);
                felix.start();
            } finally {
                Thread.currentThread().setContextClassLoader(prevCls);
            }
            deployExtensionBundle(plugins);
            BundleExtensionProvider.provideExtensions(plugins);
            installAndStartBundles(plugins);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public void stop() {
        try {
            felix.stop();
        } catch (BundleException e) {
            e.printStackTrace();
        }
    }

    public void deploy() {
        File servletTemp =
                (File) context.getAttribute("javax.servlet.context.tempdir");
        platformDirectory = new File(servletTemp, "felix");
        if (!platformDirectory.exists()) {
            platformDirectory.mkdirs();
        }
    }

    public void undeploy() {
        if (platformDirectory == null) {
            context.log("Undeploy unnecessary. - (not deployed)");
            return;
        }
        deleteDirectory(new File(platformDirectory, "plugins"));
        platformDirectory = null;
    }

    public ClassLoader getFrameworkContextClassLoader() {
        return null;  
    }

    public void installAndStartBundles(File target) {
        File[] files = target.listFiles(new Filter());
        BundleContext context = felix.getBundleContext();
        for (File file : files) {
            try {
                String toURl = file.toURL().toString();
                Bundle bundle = context.installBundle(toURl);
                //TODO: Don't start the bundles now. Find betterway to do this. Use Equinox like way.
//                bundle.start();
            } catch (MalformedURLException e) {
                e.printStackTrace();
            } catch (BundleException e) {
                e.printStackTrace();
            }
        }


    }

    private class Filter implements FileFilter {
        public boolean accept(File pathname) {
            return pathname.getName().endsWith(".jar");
        }
    }

    /**
     * deployExtensionBundle will generate the Servletbridge extensionbundle if it is not already present in the platform's
     * plugin directory. By default it exports "org.eclipse.equinox.servletbridge" and a versioned export of the Servlet API.
     * Additional exports can be added by using the "extendedFrameworkExports" initial-param in the ServletConfig
     *
     * @param plugins plugins
     */
    private void deployExtensionBundle(File plugins) throws Exception {
        File extensionBundle = new File(plugins,
                                        "org.wso2.carbon.bridge.extensionbundle_1.0.0.jar");
        Manifest mf = new Manifest();
        Attributes attribs = mf.getMainAttributes();
        attribs.putValue(MANIFEST_VERSION, "1.0");
        attribs.putValue(BUNDLE_MANIFEST_VERSION, "2");
        attribs.putValue(BUNDLE_NAME, "Servletbridge Extension Bundle");
        attribs.putValue(BUNDLE_SYMBOLIC_NAME,
                         "org.wso2.carbon.bridge");
        attribs.putValue(BUNDLE_VERSION, "1.0.0");
        attribs.putValue(FRAGMENT_HOST, "system.bundle; extension:=framework");

        String servletVersion =
                context.getMajorVersion() + "." + context.getMinorVersion();
        String packageExports = "org.wso2.carbon.bridge; version=1.0.0" +
                                ", org.wso2.carbon.server.transports; version=1.0.0" +
                                ", javax.servlet; version=" + servletVersion +
                                ", javax.servlet.http; version=" + servletVersion +
                                ", javax.servlet.resources; version=" +
                                servletVersion;

        String extendedExports = config.getInitParameter(CONFIG_EXTENDED_FRAMEWORK_EXPORTS);
                if (extendedExports != null && extendedExports.trim().length() != 0)
                    packageExports += ", " + extendedExports;
        attribs.putValue(EXPORT_PACKAGE, packageExports);

        try {
            JarOutputStream jos = null;
            try {
                jos = new JarOutputStream(new FileOutputStream(extensionBundle), mf);
                jos.finish();
            } finally {
                if (jos != null)
                    jos.close();
            }
            //install the master bundle
            felix.getBundleContext().installBundle(extensionBundle.toURL().toString());
        } catch (IOException e) {
            context.log("Error generating extension bundle", e);
            e.printStackTrace();
            throw new Exception(e);
        } catch (BundleException e) {
            e.printStackTrace();
            throw new Exception(e);
        }
    }


    public void loadSystemProperties() {
        // The system properties file is either specified by a system
        // property or it is in the same directory as the Felix JAR file.
        // Try to load it from one of these places.

        // See if the property URL was specified as a property.
        URL propURL;
        String custom = System.getProperty(SYSTEM_PROPERTIES_PROP);
        if (custom != null) {
            try {
                propURL = new URL(custom);
            }
            catch (MalformedURLException ex) {
                System.err.print("Main: " + ex);
                return;
            }
        } else {
            // Determine where the configuration directory is by figuring
            // out where felix.jar is located on the system class path.
            File confDir = null;
            String classpath = System.getProperty("java.class.path");
            int index = classpath.toLowerCase().indexOf("felix.jar");
            int start = classpath.lastIndexOf(File.pathSeparator, index) + 1;
            if (index >= start) {
                // Get the path of the felix.jar file.
                String jarLocation = classpath.substring(start, index);
                // Calculate the conf directory based on the parent
                // directory of the felix.jar directory.
                confDir = new File(
                        new File(new File(jarLocation).getAbsolutePath()).getParent(),
                        "conf");
            } else {
                // Can't figure it out so use the current directory as default.
                confDir = new File(System.getProperty("user.dir"));
            }

            try {
                propURL = new File(confDir, SYSTEM_PROPERTIES_FILE_VALUE).toURL();
            }
            catch (MalformedURLException ex) {
                System.err.print("Main: " + ex);
                return;
            }
        }

        // Read the properties file.
        Properties props = new Properties();
        InputStream is = null;
        try {
            is = propURL.openConnection().getInputStream();
            props.load(is);
            is.close();
        }
        catch (FileNotFoundException ex) {
            // Ignore file not found.
        }
        catch (Exception ex) {
            System.err.println(
                    "Main: Error loading system properties from " + propURL);
            System.err.println("Main: " + ex);
            try {
                if (is != null) is.close();
            }
            catch (IOException ex2) {
                // Nothing we can do.
            }
            return;
        }

        // Perform variable substitution on specified properties.
        for (Enumeration e = props.propertyNames(); e.hasMoreElements();) {
            String name = (String) e.nextElement();
            System.setProperty(name,
                               substVars(props.getProperty(name), name, null, null));
        }
    }

    public Properties loadConfigProperties() throws Exception {
        // Read the properties file.
        Properties props = new Properties();
        InputStream is = null;
        try {
            // Try to load config.properties.
            is = this.getClass().getClassLoader()
                    .getResourceAsStream("conf/" + CONFIG_PROPERTIES_FILE_VALUE);
            if (is == null) {
                throw new Exception("where is : " + "conf/" + CONFIG_PROPERTIES_FILE_VALUE);
            }
            props.load(is);
            is.close();
        }
        catch (Exception e) {
            throw new Exception(e);
        }
        // Perform variable substitution for system properties.
        for (Enumeration e = props.propertyNames(); e.hasMoreElements();) {
            String name = (String) e.nextElement();
            props.setProperty(name,
                              substVars(props.getProperty(name), name, null, props));
        }

        return props;
    }

    public void copySystemProperties(Properties configProps) {
        for (Enumeration e = System.getProperties().propertyNames();
             e.hasMoreElements();) {
            String key = (String) e.nextElement();
            if (key.startsWith("felix.") ||
                key.equals("org.osgi.framework.system.packages") ||
                key.equals("org.osgi.framework.bootdelegation")) {
                configProps.setProperty(key, System.getProperty(key));
            }
        }
    }

    private static final String DELIM_START = "${";
    private static final String DELIM_STOP = "}";

    /**
     * <p>
     * This method performs property variable substitution on the
     * specified value. If the specified value contains the syntax
     * <tt>${&lt;prop-name&gt;}</tt>, where <tt>&lt;prop-name&gt;</tt>
     * refers to either a configuration property or a system property,
     * then the corresponding property value is substituted for the variable
     * placeholder. Multiple variable placeholders may exist in the
     * specified value as well as nested variable placeholders, which
     * are substituted from inner most to outer most. Configuration
     * properties override system properties.
     * </p>
     *
     * @param val         The string on which to perform property substitution.
     * @param currentKey  The key of the property being evaluated used to
     *                    detect cycles.
     * @param cycleMap    Map of variable references used to detect nested cycles.
     * @param configProps Set of configuration properties.
     * @return The value of the specified string after system property substitution.
     * @throws IllegalArgumentException If there was a syntax error in the
     *                                  property placeholder syntax or a recursive variable reference.
     */
    public String substVars(String val, String currentKey,
                            Map cycleMap, Properties configProps)
            throws IllegalArgumentException {
        // If there is currently no cycle map, then create
        // one for detecting cycles for this invocation.
        if (cycleMap == null) {
            cycleMap = new HashMap();
        }

        // Put the current key in the cycle map.
        cycleMap.put(currentKey, currentKey);

        // Assume we have a value that is something like:
        // "leading ${foo.${bar}} middle ${baz} trailing"

        // Find the first ending '}' variable delimiter, which
        // will correspond to the first deepest nested variable
        // placeholder.
        int stopDelim = val.indexOf(DELIM_STOP);

        // Find the matching starting "${" variable delimiter
        // by looping until we find a start delimiter that is
        // greater than the stop delimiter we have found.
        int startDelim = val.indexOf(DELIM_START);
        while (stopDelim >= 0) {
            int idx = val.indexOf(DELIM_START, startDelim + DELIM_START.length());
            if ((idx < 0) || (idx > stopDelim)) {
                break;
            } else if (idx < stopDelim) {
                startDelim = idx;
            }
        }

        // If we do not have a start or stop delimiter, then just
        // return the existing value.
        if ((startDelim < 0) && (stopDelim < 0)) {
            return val;
        }
        // At this point, we found a stop delimiter without a start,
        // so throw an exception.
        else if (((startDelim < 0) || (startDelim > stopDelim))
                 && (stopDelim >= 0)) {
            throw new IllegalArgumentException(
                    "stop delimiter with no start delimiter: "
                    + val);
        }

        // At this point, we have found a variable placeholder so
        // we must perform a variable substitution on it.
        // Using the start and stop delimiter indices, extract
        // the first, deepest nested variable placeholder.
        String variable =
                val.substring(startDelim + DELIM_START.length(), stopDelim);

        // Verify that this is not a recursive variable reference.
        if (cycleMap.get(variable) != null) {
            throw new IllegalArgumentException(
                    "recursive variable reference: " + variable);
        }

        // Get the value of the deepest nested variable placeholder.
        // Try to configuration properties first.
        String substValue = (configProps != null)
                            ? configProps.getProperty(variable, null)
                            : null;
        if (substValue == null) {
            // Ignore unknown property values.
            substValue = System.getProperty(variable, "");
        }

        // Remove the found variable from the cycle map, since
        // it may appear more than once in the value and we don't
        // want such situations to appear as a recursive reference.
        cycleMap.remove(variable);

        // Append the leading characters, the substituted value of
        // the variable, and the trailing characters to get the new
        // value.
        val = val.substring(0, startDelim)
              + substValue
              + val.substring(stopDelim + DELIM_STOP.length(), val.length());

        // Now perform substitution again, since there could still
        // be substitutions to make.
        val = substVars(val, currentKey, cycleMap, configProps);

        // Return the value.
        return val;
    }

    /**
     * copyResource is a convenience method to recursively copy resources from the ServletContext to
     * an installation target. The default behaviour will create a directory if the resourcepath ends
     * in '/' and a file otherwise.
     *
     * @param resourcePath - The resource root path
     * @param target       - The root location where resources are to be copied
     */
    protected void copyResource(String resourcePath, File target) {
        if (resourcePath.endsWith("/")) { 
            target.mkdir();
            Set paths = context.getResourcePaths(resourcePath);
            if (paths == null)
                return;
            for (Iterator it = paths.iterator(); it.hasNext();) {
                String path = (String) it.next();
                File newFile = new File(target, path.substring(resourcePath.length()));
                copyResource(path, newFile);
            }
        } else {
            try {
                if (target.createNewFile()) {
                    InputStream is = null;
                    OutputStream os = null;
                    try {
                        is = context.getResourceAsStream(resourcePath);
                        if (is == null)
                            return;
                        os = new FileOutputStream(target);
                        byte[] buffer = new byte[8192];
                        int bytesRead = is.read(buffer);
                        while (bytesRead != -1) {
                            os.write(buffer, 0, bytesRead);
                            bytesRead = is.read(buffer);
                        }
                    } finally {
                        if (is != null)
                            is.close();

                        if (os != null)
                            os.close();
                    }
                }
            } catch (IOException e) {
                context.log("Error copying resources", e);
            }
        }
    }

    /**
     * deleteDirectory is a convenience method to recursively delete a directory
     *
     * @param directory - the directory to delete.
     * @return was the delete succesful
     */
    protected static boolean deleteDirectory(File directory) {
        if (directory.exists() && directory.isDirectory()) {
            File[] files = directory.listFiles();
            for (int i = 0; i < files.length; i++) {
                if (files[i].isDirectory()) {
                    deleteDirectory(files[i]);
                } else {
                    files[i].delete();
                }
            }
        }
        return directory.delete();
    }
}
