/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.core;

import org.apache.axis2.AxisFault;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.axis2.engine.ListenerManager;
import static org.apache.axis2.transport.TransportListener.HOST_ADDRESS;
import org.apache.axis2.transport.http.server.HttpUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.wso2.carbon.core.security.CarbonJMXAuthenticator;
import org.wso2.carbon.core.util.ParameterUtil;
import org.wso2.carbon.core.CarbonConfigurationContextFactory;
import org.wso2.carbon.utils.CarbonUtils;
import org.wso2.carbon.utils.ManagementFactory;
import org.wso2.carbon.utils.Monitor;
import org.wso2.carbon.utils.NetworkUtils;
import org.wso2.carbon.utils.ServerConfiguration;
import org.wso2.carbon.core.ServerConfigurator;
import static org.wso2.carbon.utils.ServerConstants.ADMIN_SERVICE_GROUP;
import static org.wso2.carbon.utils.ServerConstants.WEB_RESOURCE_LOCATION;
import org.wso2.carbon.utils.ServerException;
import org.wso2.carbon.registry.core.Resource;
import org.wso2.carbon.registry.core.Registry;
import org.wso2.carbon.registry.core.service.RegistryService;

import javax.management.MBeanServer;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;
import java.io.File;
import java.net.SocketException;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.UnicastRemoteObject;
import java.util.HashMap;

/**
 * This class is responsible for managing the WSO2 WSAS instance during startup.
 */
public final class ServerManager {
    private static final Log log = LogFactory.getLog(ServerManager.class);

    public String axis2RepoLocation;
    public Monitor monitor;
    private ConfigurationContext configContext;
    public String serverWorkDir;

    private ServerConfiguration serverConfig;

    private static ServerManager instance;
    private ListenerManager listenerManager;

    private java.rmi.registry.Registry rmiRegistry;
    private JMXConnectorServer jmxConnectorServer;
    
    private BundleContext context;

    public static ServerManager getInstance() {
        if (instance == null) {
            instance = new ServerManager();
        }
        return instance;
    }

    public void setBundleContext(BundleContext context) {
        this.context = context;
    }

    public BundleContext getBundleContext() {
        return this.context;
    }

    private ServerManager() {
        this.serverConfig = ServerConfiguration.getInstance();
    }

    public void start() throws ServerException {
        try {
            ServerConfigurator serverConfigurator = ServerConfigurator.getInstance();
            serverConfigurator.setBundleContext(context);
            serverConfigurator.cleanup();
            if (!serverConfigurator.isInitialized()) {
                serverConfigurator.init(axis2RepoLocation,
                                        System.getProperty(WEB_RESOURCE_LOCATION));
            }


            //TODO As a tempory solution this part is added here. But when ui bundle are seperated from the core bundles
            //TODO this should be fixed. 
            ServerConfiguration config = ServerConfiguration.getInstance();
            String type = config.getFirstProperty("Security.KeyStore.Type");
            String password = config.getFirstProperty("Security.KeyStore.Password");
            String storeFile = new File(config.getFirstProperty("Security.KeyStore.Location")).getAbsolutePath();

            System.setProperty("javax.net.ssl.trustStore", storeFile);
            System.setProperty("javax.net.ssl.trustStoreType", type);
            System.setProperty("javax.net.ssl.trustStorePassword", password);


//            persistSecurityKeyStore();

            ConfigurationContext oldContext = configContext;
            if (oldContext != null) {
                ListenerManager lm = ListenerManager.getDefaultListenerManager();
                if (lm != null) {
                    lm.stop();
                }
            }
            ServerConfigurator serverConfigurator1 = ServerConfigurator.getInstance();
            serverConfigurator1.setBundleContext(context);
            configContext =
                    CarbonConfigurationContextFactory.createNewConfigurationContext(
                            serverConfigurator1, context);

            // Disable admin services, if needed
            if (!CarbonUtils.isAdminConsoleEnabled()) {
                AxisConfiguration axisConfig = configContext.getAxisConfiguration();
                if (axisConfig.getServiceGroup(ADMIN_SERVICE_GROUP) != null) {
                    axisConfig.removeServiceGroup(ADMIN_SERVICE_GROUP);
                }
            }

            initNetworkUtils(configContext.getAxisConfiguration());

            runInitializers();

            populateConnectionProperties(configContext);

            startJMXService();

            try {
                ServerStatus.setServerRunning();
            } catch (AxisFault e) {
                String msg = "Cannot set server to running mode";
                log.error(msg, e);
            }
        } catch (Exception e) {
            String msg = "Cannot start " + serverConfig.getFirstProperty("Name");
            log.fatal(msg, e);
            throw new ServerException(msg, e);
        }
    }

    private void populateConnectionProperties(ConfigurationContext configCtx) throws Exception {
        ServiceReference reference = context.getServiceReference(RegistryService.class.getName());
        if (reference != null) {
            RegistryService registryService = (RegistryService) context.getService(reference);
            Registry registry = registryService.getSystemRegistry();
            Resource resource = registry.newResource();
            String contextRoot = configCtx.getContextRoot();
            String servicePath = configCtx.getServicePath();
            resource.setProperty("service-path", servicePath);
            resource.setProperty("context-root", contextRoot);
            String requestIP = HttpUtils.getIpAddress(configCtx.getAxisConfiguration());
            resource.setProperty("host-name", requestIP);
            registry.put(RegistryResources.CONNECTION_PROPS, resource);
            resource.discard();
        }
    }

    public void startListenerManager() {
        listenerManager = new ListenerManager();
        listenerManager.startSystem(configContext);
        ListenerManager.defaultConfigurationContext = configContext;
    }

    public void stopListenerManager() throws AxisFault {
        try {
            if (listenerManager != null) {
                listenerManager.destroy();
            }

            //we need to call this method to clean the temp files we created.
            if (configContext != null) {
                configContext.terminate();
            }
        } catch (Exception e) {
            log.error("Exception occurred while shutting down listeners", e);
        }
    }


    public ConfigurationContext getConfigContext() {
        return configContext;
    }

    public void setConfigContext(ConfigurationContext configContext) {
        this.configContext = configContext;
    }

    private void runInitializers() throws ServerException {

        String[] initializers =
                serverConfig.getProperties("ServerInitializers.Initializer");
        for (String clazzName : initializers) {
            try {
                Class clazz = context.getBundle().loadClass(clazzName);
                ServerInitializer intializer = (ServerInitializer) clazz.newInstance();
                if (log.isDebugEnabled()) {
                    log.debug("Using ServerInitializer " + intializer.getClass().getName());
                }
                intializer.init(configContext);
            } catch (Exception e) {
                throw new ServerException(e);
            }
        }
    }

    private void startJMXService() throws ServerException {
        String jmxPort = ServerConfiguration.getInstance().getFirstProperty("Ports.JMX");
        if (jmxPort != null) {
            int jmxPortInt = Integer.parseInt(jmxPort);
            MBeanServer mbs = ManagementFactory.getMBeanServer();
            try {
                try {
                    rmiRegistry = LocateRegistry.createRegistry(jmxPortInt);
                } catch (Throwable ignored) {
                    ignored.printStackTrace();
                }

                // Create an RMI connector and start it
                String jmxURL = "service:jmx:rmi:///jndi/rmi://" +
                        NetworkUtils.getLocalHostname() + ":" + jmxPortInt + "/server";
                JMXServiceURL url = new JMXServiceURL(jmxURL);

                // Security credentials are included in the env Map
                HashMap<String, CarbonJMXAuthenticator> env = new HashMap<String, CarbonJMXAuthenticator>();
                env.put(JMXConnectorServer.AUTHENTICATOR, new CarbonJMXAuthenticator());
                jmxConnectorServer =
                        JMXConnectorServerFactory.newJMXConnectorServer(url, env, mbs);
                jmxConnectorServer.start();
                log.info("JMX Service URL        : " + jmxURL);
            } catch (Exception e) {
                String msg = "Could not initialize MBean server";
                log.error(msg, e);
            }
        }
    }

    public void stopJmxService() {
        try {
            if (jmxConnectorServer != null) {
                jmxConnectorServer.stop();
                try {
                    UnicastRemoteObject.unexportObject(rmiRegistry, true); // Stop the RMI registry
                } catch (java.rmi.NoSuchObjectException ignored) {
                }
            }
        } catch (Exception e) {
            log.error("Error occurred while stopping JMXConnectorServer", e);
        }
    }

    private void initNetworkUtils(AxisConfiguration axisConfiguration)
            throws AxisFault, SocketException {
        String hostName = serverConfig.getFirstProperty("HostName");
        if (hostName != null) {
            Parameter param = axisConfiguration.getParameter(HOST_ADDRESS);
            if (param != null) {
                param.setEditable(true);
                param.setValue(hostName);
            } else {
                param = ParameterUtil.createParameter(HOST_ADDRESS, hostName);
                axisConfiguration.addParameter(param);
            }
        } else {
            Parameter param = axisConfiguration.getParameter(HOST_ADDRESS);
            if (param != null) {
                hostName = (String) param.getValue();
                log.info(HOST_ADDRESS + " has been selected from Axis2.xml.");
            }
        }
        NetworkUtils.init(hostName);
    }
}
