/**
 *  Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.wso2.carbon.mediation.utils;

import org.apache.axis2.description.Parameter;
import org.apache.axis2.AxisFault;
import org.apache.synapse.SynapseConstants;
import org.apache.synapse.config.SynapseConfiguration;
import org.apache.synapse.core.SynapseEnvironment;
import org.apache.commons.logging.Log;
import org.wso2.carbon.core.AbstractAdmin;

import java.text.MessageFormat;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

/**
 * Provides the base class for all the admin classes related to mediation (or ESB in other words)
 * This gives the access to the <code>SynapseConfiguration</code> and
 * <code>SynapseEnvironment</code>.
 *
 * This bundle is highly dependent on the Synapse framework and requires the
 * <code>org.apache.synapse.core</code> bundle at the execution time.
 */
public class AbstractMediationAdmin extends AbstractAdmin {

    private static ResourceBundle resources;

    static {
        try {
            resources = ResourceBundle.getBundle("org.wso2.carbon.mediation.utils.errors");
        } catch (MissingResourceException e) {
            throw new RuntimeException(e.getMessage());
        }
    }

    /**
     * Access utility method for the <code>SynapseConfiguration</code>
     * 
     * @return the configuration of synapse extracted from the axis configuration
     * @throws org.apache.axis2.AxisFault if the parameter <code>synapse.config</code> is not found
     * in the given <code>axis2.xml</code>
     */
    public SynapseConfiguration getSynapseConfiguration() throws AxisFault {
        Parameter synapseConfigParam
                = getAxisConfig().getParameter(SynapseConstants.SYNAPSE_CONFIG);
        if (synapseConfigParam == null) {
            throw new AxisFault(getMessage("synapse.config.notfound", new String[0]));
        }
        return (SynapseConfiguration) synapseConfigParam.getValue();
    }

    /**
     * Access utility method for the <code>SynapseEnvironment</code>
     *
     * @return the environment of synapse extracted from the axis configuration
     * @throws org.apache.axis2.AxisFault if the parameter <code>synapse.env</code> is not found
     * in the given <code>axis2.xml</code>
     */
    public SynapseEnvironment getSynapseEnvironment() throws AxisFault {
        Parameter synapseEnvParam = getAxisConfig().getParameter(
                SynapseConstants.SYNAPSE_ENV);
        if (synapseEnvParam == null) {
            throw new AxisFault(getMessage("synapse.env.notfound", new String[0]));
        }
        return (SynapseEnvironment) synapseEnvParam.getValue();
    }

    protected static String getMessage(String errorCode, Object[] args) {
        String msg;
        try {
            msg = MessageFormat.format(resources.getString(errorCode), args);
        } catch (MissingResourceException e) {
            throw new RuntimeException("Undefined '" + errorCode + "' resource property");
        }
        return msg;
    }

    /**
     * Throws <code>AxisFault</code> based on the input parameters and logs errors
     *
     * @param log the <code>Log</code> instance to be used to log the message
     * @param message the error message
     * @param e the original exception
     * @throws AxisFault 
     */
    protected void handleFault(Log log, String message, Exception e) throws AxisFault {
        if (e == null) {
            AxisFault af = new AxisFault(message);
            log.error(message, af);
            throw af;
        } else {
            message = message + " :: " + e.getMessage();
            log.error(message, e);
            throw new AxisFault(message, e);
        }
    }
}
