/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.Comment;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.dao.dataobjects.CommentDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;

import java.sql.*;
import java.util.ArrayList;
import java.util.List;

public class CommentsDAO {

    private static final Log log = LogFactory.getLog(CommentsDAO.class);

    public static String addComment(String resourceID, String userID, Comment comment)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_COMMENT (REG_RID, REG_USER_ID, REG_COMMENT_TEXT, REG_COMMENTED_TIME) " +
                "VALUES (?, ?, ?, ?)";

        long now = System.currentTimeMillis();
        PreparedStatement ps = null;
        PreparedStatement ps2 = null;
        ResultSet result = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setString(2, userID);
            ps.setString(3, comment.getText());
            ps.setTimestamp(4, new Timestamp(now));
            ps.executeUpdate();

            sql = "SELECT REG_COMMENT_ID FROM REG_COMMENT WHERE REG_RID=? AND REG_USER_ID=? AND REG_COMMENTED_TIME=?";
            ps2 = conn.prepareStatement(sql);
            ps2.setString(1, resourceID);
            ps2.setString(2, userID);
            ps2.setTimestamp(3, new Timestamp(now));
            result = ps2.executeQuery();

            String commentID = null;
            if (result.next()) {
                commentID = result.getString("REG_COMMENT_ID");
            }
            return commentID;

        } catch (SQLException e) {

            String msg = "Failed to add comment " + comment.getText() +
                    " to resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
                if (ps2 != null) {ps2.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

    }

    public static void copyComments(String path, String resourceID) throws RegistryException {

        Comment[] comments = getComments(path);
        CommentDO[] commentDOs = new CommentDO[comments.length];
        for (int i = 0; i < comments.length; i++) {
            CommentDO commentDO = new CommentDO();
            commentDO.setCommentedUser(comments[i].getUser());
            commentDO.setCommentText(comments[i].getText());
            commentDOs[i] = commentDO;
        }

        addComments(resourceID, commentDOs);
    }

    private static void addComments(String resourceID, CommentDO[] commentDOs)
            throws RegistryException {

        long now = System.currentTimeMillis();
        Timestamp currentTime = new Timestamp(now);

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_COMMENT (REG_RID, REG_USER_ID, REG_COMMENT_TEXT, REG_COMMENTED_TIME) " +
                "VALUES (?, ?, ?, ?)";

        PreparedStatement s = null;
        try {
            s = conn.prepareStatement(sql);

            for (CommentDO commentDO: commentDOs) {
                s.setString(1, resourceID);
                s.setString(2, commentDO.getCommentedUser());
                s.setString(3, commentDO.getCommentText());
                s.setTimestamp(4, currentTime);
                s.addBatch();
            }
            
            if (commentDOs.length > 0) {
                s.executeBatch();
            }
            s.clearBatch();

        } catch (SQLException e) {

            String msg = "Failed to add comments to resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public static void updateComment(long commentId, String text) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement s = null;
        try {
            String sql = "UPDATE REG_COMMENT SET REG_COMMENT_TEXT=?,REG_COMMENTED_TIME=? WHERE REG_COMMENT_ID=?";
            long now = System.currentTimeMillis();
            s = conn.prepareStatement(sql);
            s.setString(1, text);
            s.setTimestamp(2, new Timestamp(now));
            s.setLong(3, commentId);
            s.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to update the comment with ID " + commentId +
                    " with text " + text + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public static void deleteComment(long commentId) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement s = null;
        try {
            String sql = "DELETE FROM REG_COMMENT WHERE REG_COMMENT_ID=?";

            s = conn.prepareStatement(sql);
            s.setLong(1, commentId);
            s.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to delete the comment with ID " +
                    commentId + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public static Comment[] getComments(String path) throws RegistryException {

        Connection conn = Transaction.getConnection();

        List commentList = new ArrayList();
        PreparedStatement s = null;
        ResultSet results = null;
        try {
            String sql = "SELECT C.REG_COMMENT_ID, C.REG_COMMENT_TEXT, C.REG_USER_ID, C.REG_COMMENTED_TIME " +
                    "FROM REG_COMMENT C, REG_RESOURCE R WHERE R.REG_PATH=? AND R.REG_RID=C.REG_RID";

            s = conn.prepareStatement(sql);
            s.setString(1, path);

            results = s.executeQuery();
            while (results.next()) {
                Comment comment = new Comment();
                comment.setText(results.getString(DatabaseConstants.COMMENT_TEXT_FIELD));
                comment.setUser(results.getString(DatabaseConstants.COMMENTED_USER_ID_FIELD));
                comment.setTime(results.getTimestamp(DatabaseConstants.COMMENTED_TIME_FIELD));
                comment.setResourcePath(path);
                String commentPath = path + RegistryConstants.URL_SEPARATOR + "comments:" +
                        results.getInt("REG_COMMENT_ID");
                comment.setPath(commentPath);
                comment.setCommentPath(commentPath);
                comment.setParentPath(path + RegistryConstants.URL_SEPARATOR + "comments");
                commentList.add(comment);
            }

        } catch (SQLException e) {

            String msg = "Failed to get comments on resource " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return (Comment[])commentList.toArray(new Comment[commentList.size()]);
    }

    public static void removeComments(long artifactID, Connection conn) throws SQLException {

        String sql = "DELETE FROM REG_COMMENT WHERE REG_RID=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setLong(1, artifactID);
        s.executeUpdate();
        s.close();
    }

    public static Comment getComment(long commentID) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement s =  null;
        ResultSet result = null;
        try {
            String sql = "SELECT C.REG_COMMENT_ID, C.REG_COMMENT_TEXT, C.REG_USER_ID, C.REG_COMMENTED_TIME, " +
                    "R.REG_PATH FROM REG_COMMENT C, REG_RESOURCE R WHERE C.REG_COMMENT_ID=? AND C.REG_RID=R.REG_RID";

            s = conn.prepareStatement(sql);
            s.setLong(1, commentID);

            result = s.executeQuery();

            Comment comment = null;
            if (result.next()) {
                String resourcePath = result.getString(DatabaseConstants.PATH_FIELD);
                String commentPath = resourcePath + RegistryConstants.URL_SEPARATOR + "comments:" +
                        result.getInt("REG_COMMENT_ID");

                comment = new Comment();
                comment.setText(result.getString(DatabaseConstants.COMMENT_TEXT_FIELD));
                comment.setUser(result.getString(DatabaseConstants.COMMENTED_USER_ID_FIELD));
                comment.setTime(result.getTimestamp(DatabaseConstants.COMMENTED_TIME_FIELD));
                comment.setResourcePath(resourcePath);
                comment.setPath(commentPath);
                comment.setCommentPath(commentPath);
                comment.setParentPath(resourcePath + RegistryConstants.URL_SEPARATOR + "comments");
            }
            return comment;

        } catch (SQLException e) {

            String msg = "Failed to get comment with ID " + commentID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public static long[] getCommentIDs(String resourcePath, Connection conn) throws SQLException {

        String sql = "SELECT C.REG_COMMENT_ID FROM REG_COMMENT C, REG_RESOURCE A WHERE A.REG_PATH=? AND A.REG_RID=C.REG_RID";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, resourcePath);

        List commentIDList = new ArrayList();
        ResultSet results = s.executeQuery();
        while (results.next()) {
            long commentID = results.getLong(DatabaseConstants.COMMENT_ID_FIELD);
            commentIDList.add(commentID);
        }

        try {
            if (results != null) {results.close();}
            if (s != null) {s.close();}
        }catch (SQLException ex)  {
            String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
            log.error(msg,ex);
        }
          
        long[] commentIDs = new long[commentIDList.size()];
        for (int i = 0; i < commentIDs.length; i++) {
            commentIDs[i] = (Long)commentIDList.get(i);
        }
        return commentIDs;
    }

    public static String getResourcePathOfComment(long commentID) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement s = null;
        ResultSet result = null;
        try {
            String sql = "SELECT R.REG_PATH FROM REG_RESOURCE R, REG_COMMENT C " +
                    "WHERE R.REG_RID=C.REG_RID AND C.REG_COMMENT_ID=?";

            s = conn.prepareStatement(sql);
            s.setLong(1, commentID);

            result = s.executeQuery();

            String resourcePath = null;
            if (result.next()) {
                resourcePath = result.getString(DatabaseConstants.PATH_FIELD);
            }

            return resourcePath;

        } catch (SQLException e) {

            String msg = "Failed to get the resource on which the comment with ID " +
                    commentID + " is applied. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }
}
