/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import java.sql.Connection;
import java.sql.Date;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.dataobjects.RatingDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;

public class RatingsDAO {

    private static final Log log = LogFactory.getLog(RatingsDAO.class);

    public static void addRating(String resourceID, String userID, int rating)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement s = null;
        try {
            String sql = "INSERT INTO REG_RATING (REG_RID, REG_USER_ID, REG_RATING, REG_RATED_TIME) " +
                    "VALUES (?, ?, ?, ?)";

            long now = System.currentTimeMillis();

            s = conn.prepareStatement(sql);
            s.setString(1, resourceID);
            s.setString(2, userID);
            s.setInt(3, rating);
            s.setDate(4, new Date(now));

            s.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to rate resource " + resourceID +
                    " with rating " + rating + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public static void copyRatings(String fromResourceID, String toResourceId)
            throws RegistryException {
    	ResultSet result = null;
    	PreparedStatement ps = null;
        try {
            String SQL = "SELECT REG_USER_ID, REG_RATING, REG_RATED_TIME FROM REG_RATING WHERE REG_RID=?";
            ps = Transaction.getConnection().prepareStatement(SQL);
            ps.setString(1, fromResourceID);

            result = ps.executeQuery();
            while (result.next()) {
                addRating(toResourceId, result.getString(1), result.getInt(2));
            }
        } catch (SQLException e) {
            throw new RegistryException("Couldn't copy ratings", e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public static void updateRating(String resourceID, String userID, int rating)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement s = null;
        try {
            String sql = "UPDATE REG_RATING SET REG_RATING=?, REG_RATED_TIME=? WHERE REG_RID=? AND REG_USER_ID=?";

            long now = System.currentTimeMillis();

            s = conn.prepareStatement(sql);
            s.setInt(1, rating);
            s.setDate(2, new Date(now));
            s.setString(3, resourceID);
            s.setString(4, userID);

            s.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to update the rating of resource " + resourceID +
                    " by user " + userID + " to value " + rating + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public static boolean ratingExists(String resourceID, String userID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement s = null;
        ResultSet result = null;
        try {
            String sql = "SELECT R.REG_RATING_ID FROM REG_RATING R WHERE R.REG_RID=? AND R.REG_USER_ID=?";

            s = conn.prepareStatement(sql);
            s.setString(1, resourceID);
            s.setString(2, userID);

            result = s.executeQuery();
            boolean ratingExist = false;
            if (result.next()) {
                ratingExist = true;
            }

            return ratingExist;

        } catch (SQLException e) {

            String msg = "Failed to check existence of rating on resource " +
                    resourceID + " done by user " + userID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public static float getAverageRating(String path) throws RegistryException {

        Connection conn = Transaction.getConnection();

        int sumRating;
        int countRating;

        PreparedStatement s1 = null;
        ResultSet result1 = null;
        try {

            String sql1 = "SELECT SUM(RT.REG_RATING) FROM REG_RATING RT, REG_RESOURCE R " +
                    "WHERE R.REG_PATH=? AND R.REG_RID=RT.REG_RID";

            s1 = conn.prepareStatement(sql1);
            s1.setString(1, path);

            result1 = s1.executeQuery();

            sumRating = 0;
            if (result1.next()) {
                sumRating = result1.getInt(1);
            }

        } catch (SQLException e) {

            String msg = "Failed to get sum of all ratings on resource " +
                    path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result1 != null) {result1.close();}
                if (s1 != null) {s1.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        PreparedStatement s2 = null;
        ResultSet result2 = null;
        try {
            String sql2 =
                    "SELECT COUNT(RT.REG_RATING) FROM REG_RATING RT, REG_RESOURCE R " +
                            "WHERE R.REG_PATH=? AND R.REG_RID=RT.REG_RID";

            s2 = conn.prepareStatement(sql2);
            s2.setString(1, path);

            result2 = s2.executeQuery();

            countRating = 0;
            if (result2.next()) {
                countRating = result2.getInt(1);
            }

        } catch (SQLException e) {

            String msg = "Failed to get ratings count on resource " +
                    path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result2 != null) {result2.close();}
                if (s2 != null) {s2.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        float averageRating = 0;
        if (countRating > 0) {
            averageRating = (float)sumRating / countRating;
        }

        return averageRating;
    }

    public static int getRating(String path, String userName) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT RT.REG_RATING FROM REG_RATING RT, REG_RESOURCE R " +
                "WHERE R.REG_PATH=? AND RT.REG_USER_ID=? AND R.REG_RID=RT.REG_RID";

        PreparedStatement s = null;
        ResultSet result = null;
        try {
            s = conn.prepareStatement(sql);
            s.setString(1, path);
            s.setString(2, userName);

            result = s.executeQuery();

            int rating = 0;
            if (result.next()) {
                rating = result.getInt(1);
            }

            return rating;

        } catch (SQLException e) {

            String msg = "Failed to get rating on resource " + path +
                    " done by user " + userName + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public static java.util.Date getRatedTime(String path, String userName)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT RT.REG_RATED_TIME FROM REG_RATING RT, REG_RESOURCE R " +
                "WHERE R.REG_PATH=? AND RT.REG_USER_ID=? AND R.REG_RID=RT.REG_RID";

        PreparedStatement s = null;
        ResultSet result = null;
        try {
            s = conn.prepareStatement(sql);
            s.setString(1, path);
            s.setString(2, userName);

            result = s.executeQuery();

            java.util.Date ratedTime = null;
            if (result.next()) {
                ratedTime = new java.util.Date(
                        result.getTimestamp(DatabaseConstants.RATED_TIME_FIELD).getTime());
            }

            return ratedTime;

        } catch (SQLException e) {

            String msg = "Failed to get rated time of the rating done by user " +
                    userName + " on resource " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public static void removeRatings(long artifactID, Connection conn) throws SQLException {

        String sql = "DELETE FROM REG_RATING WHERE REG_RID=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setLong(1, artifactID);
        s.executeUpdate();
        s.close();
    }

    public static String[] getRatedUserNames(String resourcePath) throws RegistryException {

        Connection conn = Transaction.getConnection();

        List userNameList = new ArrayList();

        PreparedStatement s = null;
        ResultSet results = null;
        try {
            String sql = "SELECT RT.REG_USER_ID FROM REG_RATING RT, REG_RESOURCE R WHERE R.REG_PATH=? AND R.REG_RID=RT.REG_RID";

            s = conn.prepareStatement(sql);
            s.setString(1, resourcePath);

            results = s.executeQuery();
            while (results.next()) {
                String ratedUserName = results.getString(DatabaseConstants.USER_ID_FIELD);
                userNameList.add(ratedUserName);
            }

        } catch (SQLException e) {

            String msg = "Failed to users who have rated the resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        String[] userNames = new String[userNameList.size()];
        for (int i = 0; i < userNames.length; i++) {
            userNames[i] = ((String)userNameList.get(i));
        }

        return userNames;
    }

    public static RatingDO getRating(long ratingID) throws SQLException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT R.REG_PATH, RT.REG_RID, RT.REG_USER_ID, RT.REG_RATING, RT.REG_RATED_TIME " +
                "FROM REG_RATING RT, REG_RESOURCE R WHERE R.REG_RID=RT.REG_RID AND RT.REG_RATING_ID=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setLong(1, ratingID);

        ResultSet results = s.executeQuery();
        if (results.next()) {

            java.util.Date ratedTime = new java.util.Date(
                    results.getTimestamp(DatabaseConstants.RATED_TIME_FIELD).getTime());

            RatingDO ratingDAO = new RatingDO();
            ratingDAO.setRatedUserName(results.getString(DatabaseConstants.USER_ID_FIELD));
            ratingDAO.setRatedTime(ratedTime);
            ratingDAO.setRating(results.getInt(DatabaseConstants.RATING_FIELD));
            ratingDAO.setResourcePath(results.getString(DatabaseConstants.PATH_FIELD));
            ratingDAO.setResourceID(results.getString(DatabaseConstants.AID_FIELD));

            return ratingDAO;
        }

        return null;
    }
}
