/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.dataobjects.ResourceDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.session.CurrentSession;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;
import org.wso2.carbon.registry.core.utils.UUIDGenerator;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.registry.core.*;

import java.io.IOException;
import java.io.InputStream;
import java.sql.*;
import java.util.*;
import java.util.Date;

public class ResourceDAO {

    private static final Log log = LogFactory.getLog(ResourceDAO.class);

    public static final int NULL = -1;
    public static final int RESOURCE = 0;
    public static final int COLLECTION = 1;

    private AssociationDAO associationDAO;

    public ResourceDAO() {
        associationDAO = new AssociationDAO();
    }

    /**
     * Returns the resource ID (RID) of the resource refered by the given path. If a resource does
     * not exists in the given path, it returns null.
     *
     * @param path Pure path of the resource
     * @return RID of resource exists. null if the resource does not exists.
     * @throws org.wso2.carbon.registry.core.exceptions.RegistryException something went wrong
     */
    public static String getResourceID(String path) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String resourceID = null;
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            String sql = "SELECT REG_RID FROM REG_RESOURCE WHERE REG_PATH=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, path);

            result = ps.executeQuery();

            if (result.next()) {
                resourceID = result.getString("REG_RID");
            }           
        } catch (SQLException e) {
            String msg = "Failed to get ID of the resource at path " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return resourceID;
    }

    /**
     * Returns the resource ID (RID) of the resource refered by the given path. If a resource does
     * not exists in the given path, it returns null. This method is used when the registry does
     * participate in an ongoing transaction.
     *
     * @param path Pure path of the resource
     * @param conn Database connection for accessing the REG_RESOURCE table
     * @return RID of resource exists. null if the resource does not exists.
     * @throws java.sql.SQLException something went wrong
     */
    public String getResourceID(String path, Connection conn) throws SQLException {

        String resourceID = null;

        String sql = "SELECT REG_RID FROM REG_RESOURCE WHERE REG_PATH=?";

        PreparedStatement ps = conn.prepareStatement(sql);
        ps.setString(1, path);

        ResultSet result = ps.executeQuery();

        if (result.next()) {
            resourceID = result.getString("REG_RID");
        }
        try {
            if (result != null) {result.close();}
            if (ps != null) {ps.close();}
        }catch (SQLException ex)  {
            String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
            log.error(msg,ex);
        }
        return resourceID;
    }

    public String getResourcePath(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String resourcePath = null;
        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            String sql = "SELECT REG_PATH FROM REG_RESOURCE WHERE REG_RID=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            result = ps.executeQuery();

            if (result.next()) {
                resourcePath = result.getString("REG_PATH");
            }
        } catch (SQLException e) {

            String msg = "Failed to get path of the resource with ID " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return resourcePath;
    }

    /**
     * Returns the type of the resource refered by the given path.
     *
     * @param path Pure path of a resource
     * @return  ResourceDAO.RESOURCE: path referes to a resource
     *          ResourceDAO.COLLECTION: path referes to a collection
     *          ResourceDAO.NULL: resource does not exist
     * @throws org.wso2.carbon.registry.core.exceptions.RegistryException something went wrong
     */
    public int getResourceType(String path) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        int resourceType;
        try {
            String sql = "SELECT REG_COLLECTION FROM REG_RESOURCE WHERE REG_PATH=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, path);

            result = ps.executeQuery();

            resourceType = ResourceDAO.NULL;
            if (result.next()) {
                int collection = result.getInt("REG_COLLECTION");

                if (collection == 0) {
                    resourceType = ResourceDAO.RESOURCE;
                } else {
                    resourceType = ResourceDAO.COLLECTION;
                }
            }
        } catch (SQLException e) {

            String msg = "Faile to get the resource type of the resource " +
                    path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return resourceType;
    }

    public boolean resourceExists(String path) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        boolean exists;
        try {
            String sql = "SELECT REG_PATH FROM REG_RESOURCE WHERE REG_PATH=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, path);

            result = ps.executeQuery();
            exists = result.next();
        } catch (SQLException e) {
            String msg = "Failed to check the existence of the resource " +
                    path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return exists;
    }

    public boolean resourceExistsWithID(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        boolean exists;
        try {
            String sql = "SELECT REG_PATH FROM REG_RESOURCE WHERE REG_RID=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            result = ps.executeQuery();
            exists = result.next();
        } catch (SQLException e) {
            String msg = "Failed to check the existence of the resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return exists;
    }

    public void attachChild(String parentID, String childID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "INSERT INTO REG_DEPENDENCY (REG_PARENT_RID, REG_CHILD_RID) VALUES (?, ?)";
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, parentID);
            ps.setString(2, childID);

            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {
            String msg = "Failed to add child resource  " + childID +
                    " to parent resource " + parentID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void detachChild(String childID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "DELETE FROM REG_DEPENDENCY WHERE REG_CHILD_RID=?";
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, childID);

            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {
            String msg = "Failed to remove the child resource with ID " +
                    childID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    /**
     * Returns the resource in the given path filled with metadata and access to the content.
     * If a resource does not exist in the given path, null is returned.
     *
     * @param path Path of the resource.
     *
     * @return ResourceImpl filled with resource metadata and access to the resource content.
     * @throws org.wso2.carbon.registry.core.exceptions.RegistryException something went wrong
     */
    public ResourceImpl get(String path) throws RegistryException {

        ResourceImpl resourceImpl = getResource(path);

        //if (resourceImpl instanceof Collection) {
        //    fillChildren((CollectionImpl)resourceImpl, -1, -1);
        //}

        fillResourceProperties(resourceImpl);

        return resourceImpl;
    }

    public CollectionImpl get(String path, int start, int pageLen) throws RegistryException {
        CollectionImpl collection = (CollectionImpl)getResource(path);
        fillChildren(collection, start, pageLen);
        fillResourceProperties(collection);
        return collection;
    }

    public void fillResourceProperties(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            String propSQL = "SELECT REG_NAME, REG_PROPERTY_VALUE FROM REG_PROPERTY WHERE REG_RID=?";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, resourceImpl.getId());
            results = ps.executeQuery();

            while (results.next()) {
                String name = results.getString("REG_NAME");
                String value = results.getString("REG_PROPERTY_VALUE");
                resourceImpl.addProperty(name, value);
            }

        } catch (SQLException e) {
            String msg = "Failed to add properties to the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        resourceImpl.setPropertiesModified(false);
    }

    public void addRoot(String path, ResourceImpl resourceImpl)
            throws RegistryException {

        addResource(path, resourceImpl);

        addProperties(resourceImpl);
    }

    public void add(String path, String parentID, ResourceImpl resourceImpl)
            throws RegistryException {

        if (!(resourceImpl instanceof org.wso2.carbon.registry.core.Collection)) {
            addContentBytes(resourceImpl);
//            addContent(resourceImpl);
        }

        addResource(path, resourceImpl);

        addDependency(parentID, resourceImpl.getId());

        addProperties(resourceImpl);
    }

    public void update(String resourceID, ResourceImpl resourceImpl) throws RegistryException {


        if (resourceImpl.isContentModified()) {
            addContentBytes(resourceImpl);
//            addContent(resourceImpl);
        }

        if (resourceImpl.isPropertiesModified()) {
            removeProperties(resourceImpl);
        }

        updateResource(resourceID, resourceImpl);

        if (resourceImpl.isPropertiesModified()) {
            updateProperties(resourceImpl);
        }
    }

    public int getEquivalentVersion(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        int equivalentVersion = -2;
        try {
            String sql = "SELECT REG_EQUIVALENT_VERSION FROM REG_RESOURCE WHERE REG_RID=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            ResultSet result = ps.executeQuery();
            if (result.next()) {
                equivalentVersion = result.getInt("REG_EQUIVALENT_VERSION");
            }

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to get the equivalen version of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return equivalentVersion;
    }

    public void markDirty(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "UPDATE REG_RESOURCE SET REG_EQUIVALENT_VERSION=?, REG_ASSOCIATED_SNAPSHOT_ID=? " +
                    "WHERE REG_RID=?";
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setLong(1, -1);
            ps.setLong(2, -1);
            ps.setString(3, resourceID);
            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to update the equivalent version of the resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void delete(String resourceID) throws RegistryException {

        String contentID = getContentID(resourceID);

        deleteProperties(resourceID);

        deleteResource(resourceID);

        if (contentID != null) {
            // delete content checks whether the content usage is zero before deleting the content
            deleteContent(contentID);
        }

        // clear authorizations for the deleted resource, only if it is not versioned. Authorizations are required to
        // access versions of the resource, although the current version of the resource is deleted.
        if (!ResourceVersionDAO.versionExist(resourceID)) {
            AuthorizationUtils.clearAuthorizations(resourceID);
        }
    }

    private String getContentID(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            String sql = "SELECT REG_CONTENT_ID FROM REG_RESOURCE WHERE REG_RID=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            String contentID = null;
            result = ps.executeQuery();
            if (result.next()) {
                contentID = result.getString("REG_CONTENT_ID");
            }
            ps.close();

            return contentID;

        } catch (SQLException e) {

            String msg = "Failed to get content ID of the resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public void deleteSubtree(String resourceID) throws RegistryException {

        List <String> childIDs = getChildIDs(resourceID);

        Iterator <String> iChildIDs = childIDs.iterator();
        while (iChildIDs.hasNext()) {
            String childID = iChildIDs.next();
            deleteSubtree(childID);
        }

        String resourcePath = getResourcePath(resourceID);
        associationDAO.removeAllAssociations(resourcePath);

        delete(resourceID);
    }

    public List <String> getChildIDs(String resourceID) throws RegistryException {

        List <String> childIDs = new ArrayList();

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            String sql = "SELECT REG_CHILD_RID FROM REG_DEPENDENCY WHERE REG_PARENT_RID=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            results = ps.executeQuery();
            while (results.next()) {
                childIDs.add(results.getString("REG_CHILD_RID"));
            }
        } catch (SQLException e) {

            String msg = "Failed to get child resource IDs of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }


        return childIDs;
    }

    public int getChildCount(String resourceID, Connection conn) throws RegistryException {
    	PreparedStatement ps = null;
    	ResultSet results = null;
        try {
            int childCount = 0;

            String sql = "SELECT REG_CHILD_RID FROM REG_DEPENDENCY WHERE REG_PARENT_RID=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            results = ps.executeQuery();
            while (results.next()) {

                String childID = results.getString("REG_CHILD_RID");

                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {
                    childCount++;
                }
            }
            return childCount;
            
        } catch (SQLException e) {
            String msg = "Failed to get child count of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public List <String> getChildPaths(String resourceID, Connection conn)
            throws RegistryException {

        List <String> childPaths = new ArrayList <String> ();
        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            String sql = "SELECT R.REG_PATH, R.REG_RID FROM REG_RESOURCE R, REG_DEPENDENCY D WHERE " +
                    "R.REG_RID=D.REG_CHILD_RID AND REG_PARENT_RID=? ORDER BY R.REG_PATH";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            results = ps.executeQuery();
            while (results.next()) {
                String childID = results.getString("REG_RID");
                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {
                    childPaths.add(results.getString("REG_PATH"));
                }
            }
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to get child resource IDs of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return childPaths;
    }

    public List <String> getChildPaths(String resourceID, int start, int pageLen, Connection conn)
            throws RegistryException {
    	PreparedStatement ps = null;
    	ResultSet results = null;
        try {

            String sql = "SELECT R.REG_RID, R.REG_PATH  FROM REG_RESOURCE R, REG_DEPENDENCY D WHERE D.REG_PARENT_RID=?" +
                    " AND D.REG_CHILD_RID=R.REG_RID ORDER BY R.REG_PATH";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            results = ps.executeQuery();

            List<String> childPathList = new ArrayList <String> ();
            int current = 0;
            int end = start + pageLen;
            while (results.next() && current <= end) {
                String childID = results.getString("REG_RID");
                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {
                    if (current >= start && (pageLen == -1 || current < end))
                        childPathList.add(results.getString("REG_PATH"));
                    current++;
                }
            }

            if (current < start) {
                throw new RegistryException("Didn't have enough results to start at #" + start);
            }

            ps.close();

            return childPathList;

        } catch (Exception e) {

            String msg = "Failed to get child paths of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    /**
     * Return the author (the user who creates the resource) if a resource exists at given path.
     * If a resource does not exist, null is returned.
     *
     * @param path Path of the resource.
     * @return Author if the resourc exists. Null otherwiese.
     * @throws RegistryException On any database related exception.
     */
    public String getResourceAuthor(String path) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "SELECT REG_CREATOR FROM REG_RESOURCE WHERE REG_PATH=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, path);

            ResultSet result = ps.executeQuery();

            String author = null;
            if (result.next()) {
                author = result.getString("REG_CREATOR");
            }

            ps.close();

            return author;

        } catch (SQLException e) {

            String msg = "Failed to get the author of the resource " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void fillChildren(CollectionImpl collection, int start, int pageLen)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        // TODO - Cache the results in a more manageable way when using pagination!

        try {
            String sql = "SELECT R.REG_RID, R.REG_PATH FROM REG_RESOURCE R, REG_DEPENDENCY D WHERE D.REG_PARENT_RID=?" +
                    " AND D.REG_CHILD_RID=R.REG_RID ORDER BY R.REG_PATH";

            //String sql = "SELECT R.RID, R.PATH FROM REG_RESOURCE R " +
            //        "WHERE R.RID=(SELECT D.CHILD_RID FROM REG_DEPENDENCY D WHERE D.PARENT_RID=?) " +
            //        "ORDER BY R.PATH";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, collection.getId());

            List <String> childPathList = new ArrayList();
            ResultSet results = ps.executeQuery();
            int current = 0;
            int end = start + pageLen;
            while (results.next() && (pageLen == -1 || current <= end)) {
                String childID = results.getString("REG_RID");
                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {

                    if (current >= start && (pageLen == -1 || current < end)) {
                        childPathList.add(results.getString("REG_PATH"));
                    }

                    current++;
                }
            }
            if (current < start) {
                throw new RegistryException("Didn't have enough results to start at #" + start);
            }
            ps.close();

            String[] childPaths = childPathList.toArray(new String[childPathList.size()]);
            collection.setContent(childPaths);

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    collection.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private ResourceImpl getResource(String path) throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResourceImpl resourceImpl = null;
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            String sql = "SELECT REG_RID, REG_PATH, REG_MEDIA_TYPE, REG_COLLECTION, REG_CREATOR, REG_CREATED_TIME, " +
                    "REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_DESCRIPTION, REG_CONTENT_ID, " +
                    "REG_ASSOCIATED_SNAPSHOT_ID " +
                    "FROM REG_RESOURCE WHERE REG_PATH=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, path);

            result = ps.executeQuery();

            if (result.next()) {

                if (result.getInt("REG_COLLECTION") == 0) {
                    resourceImpl = new ResourceImpl();
                } else {
                    resourceImpl = new CollectionImpl();
                }

                // this is always the current version of the resource
                resourceImpl.setVersionNumber(-1);

                resourceImpl.setId(result.getString("REG_RID"));
                resourceImpl.setPath(result.getString("REG_PATH"));
                resourceImpl.setMediaType(result.getString("REG_MEDIA_TYPE"));
                resourceImpl.setAuthorUserName(result.getString("REG_CREATOR"));
                resourceImpl.setCreatedTime(result.getTimestamp("REG_CREATED_TIME"));
                resourceImpl.setLastUpdaterUserName(result.getString("REG_LAST_UPDATOR"));
                resourceImpl.setLastModified(result.getTimestamp("REG_LAST_UPDATED_TIME"));
                resourceImpl.setDescription(result.getString("REG_DESCRIPTION"));
                resourceImpl.setDbBasedContentID(result.getString("REG_CONTENT_ID"));
                resourceImpl.setMatchingSnapshotID(result.getLong("REG_ASSOCIATED_SNAPSHOT_ID"));

                if (resourceImpl.getDbBasedContentID() != null) {
                    fillResourceContent(conn, resourceImpl);
                }
            }
        } catch (SQLException e) {
            String msg = "Failed to get the resource at path " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return resourceImpl;
    }

    private void fillResourceContent(Connection conn, ResourceImpl resourceImpl) throws SQLException, RegistryException {
        String resourceContentSQL = "SELECT  REG_CONTENT_DATA  FROM  REG_CONTENT WHERE REG_CONTENT_ID = ?";
        PreparedStatement ps2  = conn.prepareStatement(resourceContentSQL);
        ps2.setString(1, resourceImpl.getDbBasedContentID());
        ResultSet result2 = ps2.executeQuery();
        if(result2.next()){
        	resourceImpl.setContentStream(
        			RegistryUtils.getMemoryStream(result2.getBinaryStream("REG_CONTENT_DATA")));
        }
        
        try {
            if (result2 != null) {result2.close();}
            if (ps2 != null) {ps2.close();}
          }
          catch (SQLException ex)  {
            String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
            log.error(msg,ex);
            throw new RegistryException(msg,ex);
          }        
    }

    public ResourceImpl getResourceByID(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            String sql = "SELECT REG_RID, REG_PATH, REG_MEDIA_TYPE, REG_COLLECTION, REG_CREATOR, REG_CREATED_TIME, " +
                    "REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_DESCRIPTION, REG_CONTENT_ID, " +
                    "REG_ASSOCIATED_SNAPSHOT_ID " +
                    "FROM REG_RESOURCE WHERE REG_RID=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            result = ps.executeQuery();

            ResourceImpl resourceImpl = null;

            if (result.next()) {

                if (result.getInt("REG_COLLECTION") == 0) {
                    resourceImpl = new ResourceImpl();
                } else {
                    resourceImpl = new CollectionImpl();
                }

                // this is always the current version of the resource
                resourceImpl.setVersionNumber(-1);

                resourceImpl.setId(result.getString("REG_RID"));
                resourceImpl.setPath(result.getString("REG_PATH"));
                resourceImpl.setMediaType(result.getString("REG_MEDIA_TYPE"));
                resourceImpl.setAuthorUserName(result.getString("REG_CREATOR"));
                resourceImpl.setCreatedTime(result.getTimestamp("REG_CREATED_TIME"));
                resourceImpl.setLastUpdaterUserName(result.getString("REG_LAST_UPDATOR"));
                resourceImpl.setLastModified(result.getTimestamp("REG_LAST_UPDATED_TIME"));
                resourceImpl.setDescription(result.getString("REG_DESCRIPTION"));
                resourceImpl.setDbBasedContentID(result.getString("REG_CONTENT_ID"));
                resourceImpl.setMatchingSnapshotID(result.getLong("REG_ASSOCIATED_SNAPSHOT_ID"));

                if (resourceImpl.getDbBasedContentID() != null) {
                    fillResourceContent(conn, resourceImpl);
                }
            }

            return resourceImpl;

        } catch (SQLException e) {

            String msg = "Failed to get the resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public InputStream getMemoryBasedStreamFromDB(String contentID, Connection conn)
            throws RegistryException {
    	PreparedStatement ps = null;
    	ResultSet result = null;
        try {
            InputStream contentStream = null;

            String sql = "SELECT REG_CONTENT_DATA FROM REG_CONTENT WHERE REG_CONTENT_ID=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);

            result = ps.executeQuery();

            if (result.next()) {
                contentStream = RegistryUtils.getMemoryStream(result.getBinaryStream("REG_CONTENT_DATA"));
            }

            return contentStream;

        } catch (SQLException e) {
            String msg = "Faild to get resource content from the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public byte[] getContentBytesFromDB(String contentID, Connection conn)
            throws RegistryException {

        try {
            byte[] contentBytes = null;

            String sql = "SELECT REG_CONTENT_DATA FROM REG_CONTENT WHERE REG_CONTENT_ID=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);

            ResultSet result = ps.executeQuery();

            if (result.next()) {
                contentBytes = result.getBytes("REG_CONTENT_DATA");
            }
            ps.close();

            return contentBytes;

        } catch (SQLException e) {
            String msg = "Faild to get resource content from the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void addResource(String path, ResourceImpl resourceImpl)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "INSERT INTO REG_RESOURCE (REG_RID, REG_PATH, REG_MEDIA_TYPE, REG_COLLECTION, REG_CREATOR, " +
                    "REG_CREATED_TIME, REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_DESCRIPTION, " +
                    "REG_CONTENT_ID, REG_EQUIVALENT_VERSION, REG_ASSOCIATED_SNAPSHOT_ID) " +
                    "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";

            long now = System.currentTimeMillis();

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceImpl.getId());
            ps.setString(2, path);
            ps.setString(3, resourceImpl.getMediaType());
            ps.setInt(4, resourceImpl instanceof org.wso2.carbon.registry.core.Collection ? 1 : 0);
            ps.setString(5, CurrentSession.getUser());
            ps.setTimestamp(6, new Timestamp(now));
            ps.setString(7, CurrentSession.getUser());
            ps.setTimestamp(8, new Timestamp(now));
            ps.setString(9, resourceImpl.getDescription());
            ps.setString(10, resourceImpl.getDbBasedContentID());
            ps.setLong(11, -1); // this is a new resource. so there is no equivalent version.

            // we are not aware of any snapshots created for this state of the resource. if a
            // snapshot is created, version repository is responsible for updating this feild.
            ps.setLong(12, -1);

            ps.executeUpdate();

            ps.close();

            // Update state of the ResourceImpl
            resourceImpl.setLastModified(new Date(now));
            resourceImpl.setPath(path);

        } catch (SQLException e) {

            String msg = "Failed to add resource to path " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void updateResource(String resourceID, ResourceImpl resourceImpl)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "UPDATE REG_RESOURCE SET REG_MEDIA_TYPE=?, REG_LAST_UPDATOR=?, " +
                    "REG_LAST_UPDATED_TIME=?, REG_DESCRIPTION=?, REG_CONTENT_ID=? WHERE REG_RID=?";

            long now = System.currentTimeMillis();

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceImpl.getMediaType());
            ps.setString(2, CurrentSession.getUser());
            ps.setTimestamp(3, new Timestamp(now));
            ps.setString(4, resourceImpl.getDescription());
            ps.setString(5, resourceImpl.getDbBasedContentID());
            ps.setString(6, resourceID);

            ps.executeUpdate();

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to update the resource at path " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void deleteResource(String resourceID) throws RegistryException {

        setParentEquivalentVersion(resourceID, -1);

        Connection conn = Transaction.getConnection();

        try {
            String sql = "DELETE FROM REG_RESOURCE WHERE REG_RID=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            ps.executeUpdate();

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to delete the resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void deleteProperties(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "DELETE FROM REG_PROPERTY WHERE REG_RID=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            ps.executeUpdate();

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to delete properties of the resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void setParentEquivalentVersion(String childID, int equivalentVersion)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "UPDATE REG_RESOURCE SET REG_EQUIVALENT_VERSION=? " +
                    "WHERE REG_RID=(SELECT D.REG_PARENT_RID FROM REG_DEPENDENCY D WHERE D.REG_CHILD_RID=?)";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setInt(1, equivalentVersion);
            ps.setString(2, childID);

            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to set the equivalent version of the parent of child resource " +
                    childID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void addDependency(String parentID, String childID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "INSERT INTO REG_DEPENDENCY (REG_PARENT_RID, REG_CHILD_RID) VALUES (?, ?)";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, parentID);
            ps.setString(2, childID);

            ps.executeUpdate();

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add the dependency between resource IDs " +
                    parentID + " and " + childID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void addProperties(Resource resource) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "INSERT INTO REG_PROPERTY (REG_RID, REG_NAME, REG_PROPERTY_VALUE) VALUES (?, ?, ?)";

            PreparedStatement ps = conn.prepareStatement(sql);

            String resourceID = resource.getId();
            Properties props = resource.getProperties();
            boolean emptyPreparedStatement = true;

            if (props != null) {
                Iterator i = props.keySet().iterator();
                while (i.hasNext()) {
                    String name = (String)i.next();

                    List propValues = (List) props.get(name);
                    if (propValues != null) {
                        Iterator<String> iValues = propValues.iterator();
                        while (iValues.hasNext()) {
                            String value = iValues.next();

                            ps.setString(1, resourceID);
                            ps.setString(2, name);
                            ps.setString(3, value);
                            ps.addBatch();
                            ps.clearParameters();
                            emptyPreparedStatement = false;
                        }
                    }
                }
                if(!emptyPreparedStatement){
                    ps.executeBatch();
                }
            }

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add properties to the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void removeProperties(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {

            String sql = "DELETE FROM REG_PROPERTY WHERE REG_RID=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceImpl.getId());

            ps.executeUpdate();

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed tp remove properties of the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void updateProperties(ResourceImpl resourceImpl) throws RegistryException {

        //removeProperties(resourceImpl);
        addProperties(resourceImpl);
    }

    private void addContentBytes(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        if (resourceImpl.getContent() == null || !(resourceImpl.getContent() instanceof byte[])) {
            if (log.isDebugEnabled()) {
                String msg = "Content of the resource " + resourceImpl.getPath() +
                        " is null or not a byte array. Content will not be persisted to the database.";
                log.debug(msg);
            }
            return;
        }

        InputStream contentStream = resourceImpl.getContentStream();
//        byte[] contentBytes = (byte[]) resourceImpl.getContent();

        try {
            String contentID = resourceImpl.getDbBasedContentID();

            String sql = "INSERT INTO REG_CONTENT (REG_CONTENT_ID, REG_CONTENT_DATA) VALUES (?, ?)";

            int size = contentStream.available();
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);
            ps.setBinaryStream(2, contentStream, size);
            ps.executeUpdate();
            ps.close();

        } catch (Exception e) {

            String msg = "Failed to write resource content to the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public String addContentBytes(byte[] contentBytes) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {

            String contentID = UUIDGenerator.generateUUID();

            String sql = "INSERT INTO REG_CONTENT (REG_CONTENT_ID, REG_CONTENT_DATA) VALUES (?, ?)";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);
            ps.setBytes(2, contentBytes);
            ps.executeUpdate();
            ps.close();

            return contentID;

        } catch (SQLException e) {

            String msg = "Failed to write resource content to the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void updateContentBytes(String contentID, byte[] contentBytes)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {

            String sql = "UPDATE REG_CONTENT SET REG_CONTENT_DATA=? WHERE REG_CONTENT_ID=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setBytes(1, contentBytes);
            ps.setString(2, contentID);
            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to write resource content to the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    /**
     * Content of resources are stored seprately in the REG_CONTENT table and given a UUID. Resource
     * refer to thier contents by this UUID. Multiple resources may refer to a same content.
     * Therefore, when deleting a resource, we can't simply delete its content. We have check if
     * the content of the deleted resource is referred by other resources. If it is referred by
     * other resources, content should NOT be deleted. If the content is not referred by any other
     * resource, we are free to delete that content.
     *
     * @param contentID ID of the content
     * @throws RegistryException On database related exceptions.
     */
    private void deleteContent(String contentID) throws RegistryException {

        if (getContentUsage(contentID) > 0) {
            return;
        }

        Connection conn = Transaction.getConnection();

        try {
            String sql = "DELETE FROM REG_CONTENT WHERE REG_CONTENT_ID=?";
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);
            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to delete content with the ID " +
                    contentID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private long getContentUsage(String contentID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT COUNT(REG_CONTENT_ID) AS REG_USAGE FROM REG_RESOURCE WHERE REG_CONTENT_ID=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);

            long usage = 0;

            ResultSet result = ps.executeQuery();
            if (result.next()) {
                usage = result.getLong("REG_USAGE");
            }
            ps.close();

            return usage;

        } catch (SQLException e) {

            String msg = "Failed to get content usage for the content ID " +
                    contentID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void setPath(String resourceID, String newPath) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "UPDATE REG_RESOURCE SET REG_PATH=? WHERE REG_RID=?";
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, newPath);
            ps.setString(2, resourceID);

            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to change the path of the resource " +
                    newPath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public long getEquivalenVersion(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "SELECT REG_EQUIVALENT_VERSION FROM REG_RESOURCE WHERE REG_RID=?";
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            ResultSet result = ps.executeQuery();
            long equivalentVersion = -1;
            if (result.next()) {
                equivalentVersion = result.getLong("REG_EQUIVALENT_VERSION");
            }
            ps.close();

            return equivalentVersion;

        } catch (SQLException e) {

            String msg = "Failed to get equivalent version of the resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public ResourceDO getResourceDO(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT R.REG_RID, R.REG_EQUIVALENT_VERSION, R.REG_PATH, R.REG_MEDIA_TYPE, R.REG_COLLECTION, " +
                "R.REG_CREATOR, R.REG_CREATED_TIME, R.REG_LAST_UPDATOR, R.REG_LAST_UPDATED_TIME, " +
                "R.REG_DESCRIPTION, R.REG_CONTENT_ID, R.REG_ASSOCIATED_SNAPSHOT_ID " +
                "FROM REG_RESOURCE R WHERE R.REG_RID=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            ResourceDO resourceDO = null;
            ResultSet result = ps.executeQuery();
            if (result.next()) {
                resourceDO = new ResourceDO();

                resourceDO.setID(result.getString("REG_RID"));
                resourceDO.setEquivalentVersion(result.getLong("REG_EQUIVALENT_VERSION"));
                resourceDO.setPath(result.getString("REG_PATH"));
                resourceDO.setMediaType(result.getString("REG_MEDIA_TYPE"));
                resourceDO.setCollection(result.getInt("REG_COLLECTION"));
                resourceDO.setAuthor(result.getString("REG_CREATOR"));
                resourceDO.setCreatedOn(result.getTimestamp("REG_CREATED_TIME"));
                resourceDO.setLastUpdator(result.getString("REG_LAST_UPDATOR"));
                resourceDO.setLastUpdatedOn(result.getTimestamp("REG_LAST_UPDATED_TIME"));
                resourceDO.setDescription(result.getString("REG_DESCRIPTION"));
                resourceDO.setContentID(result.getString("REG_CONTENT_ID"));
                resourceDO.setAssociatedSnapshotID(result.getLong("REG_ASSOCIATED_SNAPSHOT_ID"));
            }

            return resourceDO;

        } catch (SQLException e) {

            String msg = "Failed to read resource version data for resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void setSnapshotID(String resourceID, long snapshotID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "UPDATE REG_RESOURCE SET REG_ASSOCIATED_SNAPSHOT_ID=? WHERE REG_RID=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setLong(1, snapshotID);
            ps.setString(2, resourceID);
            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to set the snapshot ID for the resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

    }

    public void markDirtyByPath(String resourcePath) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "UPDATE REG_RESOURCE SET REG_EQUIVALENT_VERSION=?, REG_ASSOCIATED_SNAPSHOT_ID=? " +
                    "WHERE REG_PATH=?";
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setLong(1, -1);
            ps.setLong(2, -1);
            ps.setString(3, resourcePath);
            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to update the equivalent version of the resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void invalidateCurrentSnapshotsInBatch(String[] resourcePaths) throws RegistryException {

        String[] sqls = new String[resourcePaths.length];
        for (int i = 0; i < sqls.length; i++) {
            sqls[i] = "UPDATE REG_RESOURCE SET REG_ASSOCIATED_SNAPSHOT_ID=-1 WHERE REG_PATH='" + resourcePaths[i] + "'";
        }

        Connection conn = Transaction.getConnection();

        try {
            Statement s = conn.createStatement();
            s.clearBatch();
            for (String sqlSt: sqls) {
                s.addBatch(sqlSt);
            }
            s.executeBatch();
            s.close();

        } catch (SQLException e) {

            String msg = "Failed to invalidate the current snapshot of the resource " +
                    resourcePaths + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }
}
