/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.dataobjects.ResourceDO;
import org.wso2.carbon.registry.core.jdbc.utils.ResourceVersionHolder;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.registry.core.*;

import java.io.IOException;
import java.io.InputStream;
import java.sql.*;
import java.util.*;

public class ResourceVersionDAO {

    private static Log log = LogFactory.getLog(ResourceVersionDAO.class);

    public long addSnapshot(String snapshotRootID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_SNAPSHOT (REG_ROOT_ID) VALUES (?)";
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, snapshotRootID);
            ps.executeUpdate();
        } catch (SQLException e) {

            String msg = "Failed to add snapshot for the resource " +
                    snapshotRootID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        String sql2 = "SELECT MAX(REG_SNAPSHOT_ID) FROM REG_SNAPSHOT WHERE REG_ROOT_ID=? GROUP BY REG_ROOT_ID";
        ResultSet result = null;
        PreparedStatement ps2 = null;
        try {
            ps2 = conn.prepareStatement(sql2);
            ps2.setString(1, snapshotRootID);

            result = ps2.executeQuery();
            long snapshotID = 0;
            if (result.next()) {
                snapshotID = result.getLong(1);
            }

            return snapshotID;

        } catch (SQLException e) {

            String msg = "Failed to get the latest snapshot ID of the resource " +
                    snapshotRootID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps2 != null) {
                    ps2.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public static boolean versionExist(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT COUNT(*) FROM REG_RESOURCE_VERSION WHERE REG_RID=?";

        boolean versionExist = false;
        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            result = ps.executeQuery();
            if (result.next()) {
                if (result.getInt(1) > 0) {
                    versionExist = true;
                }
            }

        } catch (SQLException e) {

            String msg = "Failed to check whether a version exists for the resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return versionExist;
    }

    public int getLatestResourceVersionNumber(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT MAX(REG_VERSOIN) " +
                "FROM REG_RESOURCE_VERSION WHERE REG_RID=? GROUP BY REG_RID";

        int snapshotNumber = 0;
        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            result = ps.executeQuery();
            if (result.next()) {
                snapshotNumber = result.getInt(1);
            }

        } catch (SQLException e) {

            String msg = "Failed to get the latest snapshot number of the resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return snapshotNumber;
    }

    public int addResourceVersion(ResourceImpl resourceImpl, String contentVersionID)
            throws RegistryException {

        int versionNumber = getMaxResourceVersion(resourceImpl.getId()) + 1;

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_RESOURCE_VERSION (REG_RID, REG_VERSION, REG_PATH, REG_MEDIA_TYPE, REG_COLLECTION, " +
                "REG_CREATOR, REG_CREATED_TIME, REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_DESCRIPTION, REG_CONTENT_ID, " +
                "REG_ASSOCIATED_SNAPSHOT_ID) " +
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);

            ps.setString(1, resourceImpl.getId());
            ps.setInt(2, versionNumber);
            ps.setString(3, resourceImpl.getPath());
            ps.setString(4, resourceImpl.getMediaType());
            ps.setInt(5, (resourceImpl instanceof org.wso2.carbon.registry.core.Collection) ? 1 : 0);
            ps.setString(6, resourceImpl.getAuthorUserName());
            ps.setTimestamp(7, new Timestamp(resourceImpl.getCreatedTime().getTime()));
            ps.setString(8, resourceImpl.getLastUpdaterUserName());
            ps.setTimestamp(9, new Timestamp(resourceImpl.getLastModified().getTime()));
            ps.setString(10, resourceImpl.getDescription());
            ps.setString(11, contentVersionID);
            ps.setLong(12, resourceImpl.getMatchingSnapshotID());

            ps.executeUpdate();
        } catch (SQLException e) {

            String msg = "Failed to add new version for resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return versionNumber;
    }

    public int getMaxResourceVersion(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT MAX(REG_VERSION) FROM REG_RESOURCE_VERSION WHERE REG_RID=? GROUP BY REG_RID";

        int maxVersion = 0;
        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            result = ps.executeQuery();
            if (result.next()) {
                maxVersion = result.getInt(1);
            }

        } catch (SQLException e) {

            String msg = "Failed to latest version number of the resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return maxVersion;
    }

    public void addDependencies(String resourceID, int versionNumber, List<String> childIDs)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_DEPENDENCY_VERSION (REG_PARENT_RID, REG_PARENT_VERSION, REG_CHILD_RID) " +
                "VALUES (?, ?, ?)";
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);

            Iterator<String> iChildIDs = childIDs.iterator();
            while (iChildIDs.hasNext()) {
                String childID = iChildIDs.next();

                ps.setString(1, resourceID);
                ps.setInt(2, versionNumber);
                ps.setString(3, childID);
                ps.addBatch();
                ps.clearParameters();
            }
            if (childIDs.size() > 0) {
                ps.executeBatch();
            }

        } catch (SQLException e) {

            String msg = "Failed to add dependencies for version " + versionNumber +
                    " of resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.clearBatch();
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public void addProperties(ResourceImpl resource, long version) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql = "INSERT INTO REG_PROPERTY_VERSION (REG_RID, REG_VERSION, REG_NAME, REG_PROPERTY_VALUE) " +
                    "VALUES (?, ?, ?, ?)";

            ps = conn.prepareStatement(sql);

            String resourceID = resource.getId();
            Properties props = resource.getProperties();
            boolean emptyPreparedStatement = true;
            if (props != null) {

                Iterator i = props.keySet().iterator();
                while (i.hasNext()) {
                    String name = (String) i.next();

                    List propValues = (List) props.get(name);
                    if (propValues != null) {
                        Iterator<String> iValues = propValues.iterator();
                        while (iValues.hasNext()) {
                            String value = iValues.next();

                            ps.setString(1, resourceID);
                            ps.setLong(2, version);
                            ps.setString(3, name);
                            ps.setString(4, value);
                            ps.addBatch();
                            emptyPreparedStatement = false;
                            ps.clearParameters();
                        }
                    }
                }
                if (!emptyPreparedStatement) {
                    ps.executeBatch();
                }

            }

        } catch (SQLException e) {

            String msg = "Failed to version properties to the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.clearBatch();
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public void addSnapshotResource(long snapshotID, String resourceID, int versionNumber)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_SNAPSHOT_RESOURCE_VERSION (REG_SNAPSHOT_ID, REG_RID, REG_VERSION) " +
                "VALUES (?, ?, ?)";
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setLong(1, snapshotID);
            ps.setString(2, resourceID);
            ps.setInt(3, versionNumber);

            ps.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to add snapshot resource version for version " +
                    versionNumber + " of resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public Long[] getSnapshotIDs(String resourcePath) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT REG_SNAPSHOT_ID FROM REG_SNAPSHOT " +
                "WHERE REG_ROOT_ID=(SELECT REG_RID FROM REG_RESOURCE WHERE REG_PATH=?) ORDER BY REG_SNAPSHOT_ID DESC";

        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, resourcePath);

            results = ps.executeQuery();
            List<Long> snapshotIDs = new ArrayList();
            while (results.next()) {
                long snapshotNumber = results.getLong("REG_SNAPSHOT_ID");
                snapshotIDs.add(snapshotNumber);
            }

            return snapshotIDs.toArray(new Long[snapshotIDs.size()]);

        } catch (SQLException e) {

            String msg = "Failed to get snapshot numbers of resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public long getVersion(String resourceID, long snapshotID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT REG_VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION WHERE REG_SNAPSHOT_ID=? AND REG_RID=?";

        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setLong(1, snapshotID);
            ps.setString(2, resourceID);

            result = ps.executeQuery();
            long version = -1;
            if (result.next()) {
                version = result.getLong("REG_VERSION");
            }

            return version;

        } catch (SQLException e) {

            String msg = "Failed to get version of resource " + resourceID +
                    " of snapshot " + snapshotID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public ResourceImpl get(String resourceID, long snapshotID, long version)
            throws RegistryException {

        ResourceImpl resourceImpl = getResource(resourceID, version);

        //if (resourceImpl instanceof Collection) {
        //    fillChildren(resourceImpl, snapshotID);
        //}

        fillResourceProperties(resourceImpl);

        return resourceImpl;
    }

    public CollectionImpl get(String resourceID, long snapshotID,
                              long version, int start, int pageLen)
            throws RegistryException {

        ResourceImpl resourceImpl = getResource(resourceID, version);

        if (!(resourceImpl instanceof CollectionImpl)) {
            String msg = "Child resource range can only be specified for collections. " +
                    resourceImpl.getPath() + " is not a collection.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        CollectionImpl collectionImpl = (CollectionImpl) resourceImpl;

        fillChildren(collectionImpl, snapshotID, start, pageLen);

        fillResourceProperties(collectionImpl);

        return collectionImpl;
    }

    private void fillResourceProperties(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            String propSQL = "SELECT REG_NAME, REG_PROPERTY_VALUE FROM REG_PROPERTY_VERSION WHERE REG_RID=? AND REG_VERSION=?";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, resourceImpl.getId());
            ps.setLong(2, resourceImpl.getVersionNumber());
            results = ps.executeQuery();

            while (results.next()) {
                String name = results.getString("REG_NAME");
                String value = results.getString("REG_PROPERTY_VALUE");
                resourceImpl.addProperty(name, value);
            }

        } catch (SQLException e) {

            String msg = "Failed to add properties to the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        resourceImpl.setPropertiesModified(false);
    }

    public int getChildCount(String resourceID, long versionNumber, Connection conn)
            throws RegistryException {

        ResultSet results = null;
        PreparedStatement ps = null;
        try {

            int childCount = 0;

            String sql = "SELECT REG_CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE REG_PARENT_RID=? AND REG_PARENT_VERSION=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, versionNumber);

            results = ps.executeQuery();
            while (results.next()) {

                String childID = results.getString("REG_CHILD_RID");

                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {
                    childCount++;
                }
            }

            return childCount;

        } catch (SQLException e) {

            String msg = "Failed to get child count of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    private void fillChildren(ResourceImpl resourceImpl, long snapshotID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            //String sql = "SELECT R.RID, R.PATH FROM REG_RESOURCE_VERSION R, REG_DEPENDENCY_VERSION D " +
            //        "WHERE D.PARENT_RID=? AND PARENT_VERSION=? AND D.CHILD_RID=R.RID AND R.VERSION=?";

            String sql = "SELECT REG_CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE REG_PARENT_RID=? AND REG_PARENT_VERSION=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceImpl.getId());
            ps.setLong(2, resourceImpl.getVersionNumber());

            List<String> childPathList = new ArrayList();
            results = ps.executeQuery();
            while (results.next()) {

                String childID = results.getString("REG_CHILD_RID");

                String sql2 = "SELECT R.REG_PATH FROM REG_RESOURCE_VERSION R " +
                        "WHERE R.REG_RID=? AND R.REG_VERSION=(" +
                        "SELECT S.REG_VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION S " +
                        "WHERE S.REG_SNAPSHOT_ID=? AND S.REG_RID=?)";

                PreparedStatement ps2 = conn.prepareStatement(sql2);
                ps2.setString(1, childID);
                ps2.setLong(2, snapshotID);
                ps2.setString(3, childID);

                ResultSet result2 = ps2.executeQuery();
                String path = "";
                if (result2.next()) {
                    path = result2.getString("REG_PATH");
                }

                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {
                    String childPath = path + ";version:" + snapshotID;
                    childPathList.add(childPath);
                }

                try {
                    if (result2 != null) {
                        result2.close();
                    }
                    if (ps != null) {
                        ps.close();
                    }
                }
                catch (SQLException ex) {
                    String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                    log.error(msg, ex);
                    throw new RegistryException(msg, ex);
                }
            }

            String[] childPaths = childPathList.toArray(new String[childPathList.size()]);
            resourceImpl.setContent(childPaths);

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public List<String> getChildPaths(
            String resourceID, long snapshotID, long versionNumber, Connection conn)
            throws RegistryException {

        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            //String sql = "SELECT R.RID, R.PATH FROM REG_RESOURCE_VERSION R, REG_DEPENDENCY_VERSION D " +
            //        "WHERE D.PARENT_RID=? AND PARENT_VERSION=? AND D.CHILD_RID=R.RID AND R.VERSION=?";

            String sql = "SELECT REG_CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE REG_PARENT_RID=? AND REG_PARENT_VERSION=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, versionNumber);

            List<String> childPathList = new ArrayList<String>();
            results = ps.executeQuery();
            while (results.next()) {

                String childID = results.getString("REG_CHILD_RID");

                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {

                    String sql2 = "SELECT R.REG_PATH FROM REG_RESOURCE_VERSION R " +
                            "WHERE R.REG_RID=? AND R.REG_VERSION=(" +
                            "SELECT S.REG_VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION S " +
                            "WHERE S.REG_SNAPSHOT_ID=? AND S.REG_RID=?)";

                    PreparedStatement ps2 = conn.prepareStatement(sql2);
                    ps2.setString(1, childID);
                    ps2.setLong(2, snapshotID);
                    ps2.setString(3, childID);

                    ResultSet result2 = ps2.executeQuery();
                    String path = "";
                    if (result2.next()) {
                        path = result2.getString("REG_PATH");
                    }

                    String childPath = path + ";version:" + snapshotID;
                    childPathList.add(childPath);
                    try {
                        if (result2 != null) {
                            result2.close();
                        }
                        if (ps2 != null) {
                            ps2.close();
                        }
                    }
                    catch (SQLException ex) {
                        String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                        log.error(msg, ex);
                        throw new RegistryException(msg, ex);
                    }
                }

            }

            return childPathList;

        } catch (SQLException e) {

            String msg = "Failed to get child paths of snapshot " + snapshotID + " of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    private void fillChildren(ResourceImpl resourceImpl, long snapshotID, int start, int pageLen)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            //String sql = "SELECT R.RID, R.PATH FROM REG_RESOURCE_VERSION R, REG_DEPENDENCY_VERSION D " +
            //        "WHERE D.PARENT_RID=? AND PARENT_VERSION=? AND D.CHILD_RID=R.RID AND R.VERSION=?";

            String sql = "SELECT REG_CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE REG_PARENT_RID=? AND REG_PARENT_VERSION=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceImpl.getId());
            ps.setLong(2, resourceImpl.getVersionNumber());

            List<String> childPathList = new ArrayList();
            results = ps.executeQuery();

            int resultCount = 0;
            int end = start + pageLen;
            while (results.next() && resultCount <= end) {

                String childID = results.getString("REG_CHILD_RID");

                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {

                    if (resultCount >= start && resultCount < end) {

                        String sql2 = "SELECT R.REG_PATH FROM REG_RESOURCE_VERSION R " +
                                "WHERE R.REG_RID=? AND R.REG_VERSION=(" +
                                "SELECT S.REG_VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION S " +
                                "WHERE S.REG_SNAPSHOT_ID=? AND S.REG_RID=?)";

                        PreparedStatement ps2 = conn.prepareStatement(sql2);
                        ps2.setString(1, childID);
                        ps2.setLong(2, snapshotID);
                        ps2.setString(3, childID);

                        ResultSet result2 = ps2.executeQuery();
                        String path = "";
                        if (result2.next()) {
                            path = result2.getString("REG_PATH");
                        }

                        String childPath = path + ";version:" + snapshotID;
                        childPathList.add(childPath);
                        try {
                            if (result2 != null) {
                                result2.close();
                            }
                            if (ps2 != null) {
                                ps2.close();
                            }
                        }
                        catch (SQLException ex) {
                            String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                            log.error(msg, ex);
                            throw new RegistryException(msg, ex);
                        }
                    }

                    resultCount++;
                }
            }

            if (resultCount < start) {
                String msg = "Not enough child resources found to start at the child number " +
                        start + " of version " + snapshotID + " of resource " +
                        resourceImpl.getPath();
                log.error(msg);
                throw new RegistryException(msg);
            }

            String[] childPaths = childPathList.toArray(new String[childPathList.size()]);
            resourceImpl.setContent(childPaths);

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public List<String> getChildPaths(
            ResourceImpl resourceImpl, long snapshotID, int start, int pageLen, Connection conn)
            throws RegistryException {

        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            //String sql = "SELECT R.RID, R.PATH FROM REG_RESOURCE_VERSION R, REG_DEPENDENCY_VERSION D " +
            //        "WHERE D.PARENT_RID=? AND PARENT_VERSION=? AND D.CHILD_RID=R.RID AND R.VERSION=?";

            String sql = "SELECT REG_CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE REG_PARENT_RID=? AND REG_PARENT_VERSION=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceImpl.getId());
            ps.setLong(2, resourceImpl.getVersionNumber());

            List<String> childPathList = new ArrayList();
            results = ps.executeQuery();

            int resultCount = 0;
            int end = start + pageLen;
            while (results.next() && resultCount <= end) {

                String childID = results.getString("REG_CHILD_RID");

                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {

                    if (resultCount >= start && resultCount < end) {

                        String sql2 = "SELECT R.REG_PATH FROM REG_RESOURCE_VERSION R " +
                                "WHERE R.REG_RID=? AND R.REG_VERSION=(" +
                                "SELECT S.REG_VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION S " +
                                "WHERE S.REG_SNAPSHOT_ID=? AND S.REG_RID=?)";

                        PreparedStatement ps2 = conn.prepareStatement(sql2);
                        ps2.setString(1, childID);
                        ps2.setLong(2, snapshotID);
                        ps2.setString(3, childID);

                        ResultSet result2 = ps2.executeQuery();
                        String path = "";
                        if (result2.next()) {
                            path = result2.getString("REG_PATH");
                        }

                        String childPath = path + ";version:" + snapshotID;
                        childPathList.add(childPath);
                        try {
                            if (result2 != null) {
                                result2.close();
                            }
                            if (ps2 != null) {
                                ps2.close();
                            }
                        }
                        catch (SQLException ex) {
                            String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                            log.error(msg, ex);
                            throw new RegistryException(msg, ex);
                        }
                    }

                    resultCount++;
                }
            }

            if (resultCount < start) {
                String msg = "Not enough child resources found to start at the child number " +
                        start + " of version " + snapshotID + " of resource " +
                        resourceImpl.getPath();
                log.error(msg);
                throw new RegistryException(msg);
            }

            return childPathList;

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    private ResourceImpl getResource(String resourceID, long version) throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            String sql = "SELECT REG_RID, REG_PATH, REG_MEDIA_TYPE, REG_COLLECTION, REG_CREATOR, REG_CREATED_TIME, " +
                    "REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_DESCRIPTION, REG_CONTENT_ID " +
                    "FROM REG_RESOURCE_VERSION WHERE REG_RID=? AND REG_VERSION=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, version);

            result = ps.executeQuery();

            ResourceImpl resourceImpl = null;

            if (result.next()) {

                if (result.getInt("REG_COLLECTION") == 0) {
                    resourceImpl = new ResourceImpl();
                } else {
                    resourceImpl = new CollectionVersionImpl();
                }

                // this is always the current version of the resource
                resourceImpl.setVersionNumber(version);

                resourceImpl.setId(result.getString("REG_RID"));
                resourceImpl.setPath(result.getString("REG_PATH"));
                resourceImpl.setMediaType(result.getString("REG_MEDIA_TYPE"));
                resourceImpl.setAuthorUserName(result.getString("REG_CREATOR"));
                resourceImpl.setCreatedTime(result.getTimestamp("REG_CREATED_TIME"));
                resourceImpl.setLastUpdaterUserName(result.getString("REG_LAST_UPDATOR"));
                resourceImpl.setLastModified(result.getTimestamp("REG_LAST_UPDATED_TIME"));
                resourceImpl.setDescription(result.getString("REG_DESCRIPTION"));
                resourceImpl.setDbBasedContentID(result.getString("REG_CONTENT_ID"));

                if (resourceImpl.getDbBasedContentID() != null) {
                    fillResourceContent(conn, resourceImpl);
                }
            }

            return resourceImpl;

        } catch (SQLException e) {

            String msg = "Failed to get resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    private void fillResourceContent(Connection conn, ResourceImpl resourceImpl) throws SQLException, RegistryException {
        String resourceContentSQL = "SELECT REG_CONTENT_DATA FROM REG_CONTENT_VERSION WHERE REG_CONTENT_VERSION_ID=?";
        PreparedStatement ps2 = conn.prepareStatement(resourceContentSQL);
        ps2.setString(1, resourceImpl.getDbBasedContentID());
        ResultSet result2 = ps2.executeQuery();
        if (result2.next()) {
            resourceImpl.setContentStream(
                    RegistryUtils.getMemoryStream(result2.getBinaryStream("REG_CONTENT_DATA")));
        }
        try {
            if (result2 != null) {
                result2.close();
            }
            if (ps2 != null) {
                ps2.close();
            }
        }
        catch (SQLException ex) {
            String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
            log.error(msg, ex);
            throw new RegistryException(msg, ex);
        }
    }

    public String addContentVersionAsBytes(byte[] contentBytes) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String contentID = null;

        if (contentBytes != null) {

            contentID = UUID.randomUUID().toString();

            PreparedStatement ps = null;
            try {

                String sql = "INSERT INTO REG_CONTENT_VERSION " +
                        "(REG_CONTENT_VERSION_ID, REG_CONTENT_DATA) VALUES (?, ?)";

                ps = conn.prepareStatement(sql);
                ps.setString(1, contentID);
                ps.setBytes(2, contentBytes);
                ps.executeUpdate();

            } catch (SQLException e) {

                String msg = "Failed to write resource content to the database. " + e.getMessage();
                log.error(msg, e);
                throw new RegistryException(msg, e);

            } finally {
                try {
                    if (ps != null) {
                        ps.close();
                    }
                }
                catch (SQLException ex) {
                    String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                    log.error(msg, ex);
                    throw new RegistryException(msg, ex);
                }
            }
        }

        return contentID;
    }

    public byte[] getContentBytesFromDB(String contentID, Connection conn)
            throws RegistryException {

        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            byte[] contentBytes = null;

            String sql = "SELECT REG_CONTENT_DATA FROM REG_CONTENT_VERSION WHERE REG_CONTENT_VERSION_ID=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);

            result = ps.executeQuery();

            if (result.next()) {
                contentBytes = result.getBytes("REG_CONTENT_DATA");
            }

            return contentBytes;

        } catch (SQLException e) {
            String msg = "Faild to get resource content from the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public byte[] getResourceContentBytes(String contentID, Connection conn)
            throws RegistryException {

        byte[] contentBytes = null;

        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            String sql = "SELECT REG_CONTENT_DATA FROM REG_CONTENT_VERSION WHERE REG_CONTENT_VERSION_ID=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);

            result = ps.executeQuery();
            if (result.next()) {
                contentBytes = result.getBytes("REG_CONTENT_DATA");
            }

        } catch (SQLException e) {
            String msg = "Faild to get resource content from the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);

        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return contentBytes;
    }

    public ResourceDO getResourceVersionDO(String resourceID, long snapshotID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT R.REG_RID, R.REG_VERSION, R.REG_PATH, R.REG_MEDIA_TYPE, R.REG_COLLECTION, " +
                "R.REG_CREATOR, R.REG_CREATED_TIME, R.REG_LAST_UPDATOR, R.REG_LAST_UPDATED_TIME, " +
                "R.REG_DESCRIPTION, R.REG_CONTENT_ID, R.REG_ASSOCIATED_SNAPSHOT_ID " +
                "FROM REG_RESOURCE_VERSION R WHERE R.REG_RID=? AND R.REG_VERSION=" +
                "(SELECT S.REG_VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION S " +
                "WHERE S.REG_SNAPSHOT_ID=? AND S.REG_RID=?)";

        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, snapshotID);
            ps.setString(3, resourceID);

            ResourceDO resourceVersionDO = null;
            result = ps.executeQuery();
            if (result.next()) {
                resourceVersionDO = new ResourceDO();

                resourceVersionDO.setID(result.getString("REG_RID"));
                resourceVersionDO.setVersion(result.getLong("REG_VERSION"));
                resourceVersionDO.setPath(result.getString("REG_PATH"));
                resourceVersionDO.setMediaType(result.getString("REG_MEDIA_TYPE"));
                resourceVersionDO.setCollection(result.getInt("REG_COLLECTION"));
                resourceVersionDO.setAuthor(result.getString("REG_CREATOR"));
                resourceVersionDO.setCreatedOn(result.getTimestamp("REG_CREATED_TIME"));
                resourceVersionDO.setLastUpdator(result.getString("REG_LAST_UPDATOR"));
                resourceVersionDO.setLastUpdatedOn(result.getTimestamp("REG_LAST_UPDATED_TIME"));
                resourceVersionDO.setDescription(result.getString("REG_DESCRIPTION"));
                resourceVersionDO.setContentID(result.getString("REG_CONTENT_ID"));
                resourceVersionDO.setAssociatedSnapshotID(result.getLong("REG_ASSOCIATED_SNAPSHOT_ID"));
            }

            return resourceVersionDO;

        } catch (SQLException e) {

            String msg = "Failed to read resource version data for resource " +
                    resourceID + " for snapshot ID " + snapshotID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public List<String> getChildIDs(String resourceID, long version) throws RegistryException {

        List<String> childIDs = new ArrayList();

        Connection conn = Transaction.getConnection();

        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            String sql = "SELECT REG_CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE REG_PARENT_RID=? AND REG_PARENT_VERSION=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, version);

            results = ps.executeQuery();
            while (results.next()) {
                childIDs.add(results.getString("REG_CHILD_RID"));
            }

        } catch (SQLException e) {

            String msg = "Failed to get child resource IDs of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return childIDs;
    }

    /**
     * Given a resource ID and a version number there may or may not be a snapshot ID. That is,
     * whenever a snapshot is created, an associating version of the resource is prepared. But,
     * when creating a snapshot of a collection, new versions will be created for all dirty child
     * resources/collections. Such resources/collection versions do not have a matching snapshot ID.
     * Such versions are just a part of a snapshot of another collection. And it is possible to
     * have more than one matching snapshot for a given resource ID and a version. A simple
     * example is creating two snapshots of a resource, without modifying the resource.
     * <p/>
     * This method returns the latest (highest) snapshot ID for the given resource ID and
     * the version. If a mathcing snapshot is not found, it will return -1.
     *
     * @param resourceID UUID of the resource
     * @param version    Version number of the resource.
     * @return Matching snapshot number if available. -1 otherwise.
     * @throws RegistryException something went wrong
     */
    public long getMatchingSnapshotID(String resourceID, long version) throws RegistryException {

        List<Long> actualSnapshotIDs = getSnapshotIDsForResourceID(resourceID);
        if (actualSnapshotIDs.size() == 0) {
            return -1;
        }

        List<Long> possibleSnapshotIDs = getSnapshotIDs(resourceID, version);

        long matchingSID = -1;
        for (int i = 0; i < actualSnapshotIDs.size(); i++) {
            long actualSID = actualSnapshotIDs.get(i);
            if (possibleSnapshotIDs.contains(actualSID)) {
                matchingSID = actualSID;
            }
        }

        return matchingSID;
    }

    /**
     * For a given snapshot, there can be at most one resource in a given path. But there may be
     * np resources for a given path or the given path may not be included in a given snapshot.
     * Therefore, given a path and snapshot ID we should get an unique resource or no resources
     * at all.
     *
     * @param path       Path of the resource, at the time where snapshot is created
     * @param snapshotID Snapshot ID
     * @return Resource ID of the resource at given path in given snapshot, if a resource exists
     *         at given path. Null otherwise.
     * @throws RegistryException For database related errors
     */
    public String getResourceID(String path, long snapshotID) throws RegistryException {

        List<ResourceVersionHolder> pathMatches = getResourceVersionsForPath(path);
        List<ResourceVersionHolder> sidMatches = getResourceVersionsForSnapshotID(snapshotID);

        Iterator<ResourceVersionHolder> pathMatchIterator = pathMatches.iterator();
        while (pathMatchIterator.hasNext()) {
            ResourceVersionHolder rsv = pathMatchIterator.next();
            if (sidMatches.contains(rsv)) {
                return rsv.getResourceID();
            }
        }

        return null;
    }

    public void removeSnapshotsOfResource(String resourceID) throws RegistryException {

        invalidateAssociatedSnapshot(resourceID);

        Connection conn = Transaction.getConnection();

        String sql = "DELETE FROM REG_SNAPSHOT WHERE REG_ROOT_ID=?";

        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to remove snapshots of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public void invalidateAssociatedSnapshot(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "UPDATE REG_RESOURCE SET REG_ASSOCIATED_SNAPSHOT_ID=? WHERE REG_RID=?";

        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setLong(1, -1);
            ps.setString(2, resourceID);
            ps.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to invalidate associated snapshot of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    private List<ResourceVersionHolder> getResourceVersionsForPath(String path)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT REG_RID, REG_VERSION FROM REG_RESOURCE_VERSION WHERE REG_PATH=?";

        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, path);

            List<ResourceVersionHolder> rsvs = new ArrayList<ResourceVersionHolder>();
            results = ps.executeQuery();
            while (results.next()) {
                String resourceID = results.getString("REG_RID");
                long version = results.getLong("REG_VERSION");
                ResourceVersionHolder rsv = new ResourceVersionHolder(resourceID, version);
                rsvs.add(rsv);
            }

            return rsvs;

        } catch (SQLException e) {

            String msg = "Failed to get resource ID and version for resource path " +
                    path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    private List<ResourceVersionHolder> getResourceVersionsForSnapshotID(long snapshotID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT REG_RID, REG_VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION WHERE REG_SNAPSHOT_ID=?";

        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setLong(1, snapshotID);

            List<ResourceVersionHolder> rsvs = new ArrayList<ResourceVersionHolder>();
            results = ps.executeQuery();
            while (results.next()) {
                String resourceID = results.getString("REG_RID");
                long version = results.getLong("REG_VERSION");
                ResourceVersionHolder rsv = new ResourceVersionHolder(resourceID, version);
                rsvs.add(rsv);
            }

            return rsvs;

        } catch (SQLException e) {

            String msg = "Failed to get resource ID and version for snapshot ID " +
                    snapshotID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    private List<Long> getSnapshotIDsForResourceID(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT REG_SNAPSHOT_ID FROM REG_SNAPSHOT WHERE REG_ROOT_ID=? ORDER BY REG_SNAPSHOT_ID DESC";

        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            results = ps.executeQuery();
            List<Long> snapshotIDs = new ArrayList<Long>();
            while (results.next()) {
                long snapshotNumber = results.getLong("REG_SNAPSHOT_ID");
                snapshotIDs.add(snapshotNumber);
            }

            return snapshotIDs;

        } catch (SQLException e) {

            String msg = "Failed to get snapshot numbers of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

    }

    private List<Long> getSnapshotIDs(String resourceID, long version) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT REG_SNAPSHOT_ID FROM REG_SNAPSHOT_RESOURCE_VERSION WHERE REG_RID=? AND REG_VERSION=?";

        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, version);

            results = ps.executeQuery();

            List<Long> sids = new ArrayList<Long>();
            while (results.next()) {
                sids.add(results.getLong("REG_SNAPSHOT_ID"));
            }

            return sids;

        } catch (SQLException e) {

            String msg = "Failed to matching snapshot IDs for resource ID " +
                    resourceID + " and version " + version + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

    }
}
