/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.Tag;
import org.wso2.carbon.registry.core.TaggedResourcePath;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.dataobjects.TaggingDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;

import java.sql.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class TagsDAO {

    private static final Log log = LogFactory.getLog(TagsDAO.class);

    public static void addTagging(String tagName, String resourceID, String userID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement s = null;
        try {
            String sql = "INSERT INTO REG_TAG (REG_TAG_NAME, REG_RID, REG_USER_ID, REG_TAGGED_TIME) VALUES (?,?,?,?)";

            long now = System.currentTimeMillis();

            s = conn.prepareStatement(sql);
            s.setString(1, tagName);
            s.setString(2, resourceID);
            s.setString(3, userID);
            s.setDate(4, new Date(now));

            s.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to add tag " + tagName + " to resource " + resourceID +
                    " by user " + userID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public static void addTaggings(String resourceID, TaggingDO[] taggingDOs)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        long now = System.currentTimeMillis();
        Date taggedTime = new Date(now);
        PreparedStatement s = null;
        try {
            String sql = "INSERT INTO REG_TAG (REG_TAG_NAME, REG_RID, REG_USER_ID, REG_TAGGED_TIME) VALUES (?,?,?,?)";

            s = conn.prepareStatement(sql);

            for (TaggingDO taggingDO : taggingDOs) {
                s.setString(1, taggingDO.getTagName());
                s.setString(2, resourceID);
                s.setString(3, taggingDO.getTaggedUserName());
                s.setDate(4, taggedTime);
                s.addBatch();
            }

            if (taggingDOs.length > 0) {
                s.executeBatch();
            }

        } catch (SQLException e) {

            String msg = "Failed to add tags to resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (s != null) {
                    s.clearBatch();
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public static void copyTags(String fromResourceID, String toResourceID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        List<TaggingDO> tagList = new ArrayList<TaggingDO>();

        ResultSet result = null;
        PreparedStatement s = null;
        try {
            String SQL = "SELECT REG_TAG_NAME, REG_USER_ID, REG_TAGGED_TIME FROM REG_TAG WHERE REG_RID=?";
            s = conn.prepareStatement(SQL);
            s.setString(1, fromResourceID);
            result = s.executeQuery();
            while (result.next()) {
                TaggingDO taggingDO = new TaggingDO();
                taggingDO.setTagName(result.getString(1));
                taggingDO.setTaggedUserName(result.getString(2));
                taggingDO.setTaggedTime(result.getDate(3));
                tagList.add(taggingDO);
            }

        } catch (SQLException e) {
            String msg = "Failed to get tags of resource " + fromResourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }


        addTaggings(toResourceID, tagList.toArray(new TaggingDO[tagList.size()]));
    }

    public static boolean taggingExists(String tagName, String resourceID, String userID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql =
                "SELECT REG_TAG_ID FROM REG_TAG WHERE lower(REG_TAG_NAME)=lower(?) AND REG_RID=? AND REG_USER_ID=?";
        ResultSet result = null;
        PreparedStatement s = null;

        try {
            s = conn.prepareStatement(sql);
            s.setString(1, tagName);
            s.setString(2, resourceID);
            s.setString(3, userID);

            result = s.executeQuery();

            boolean tagExists = false;
            if (result.next()) {
                tagExists = true;
            }

            return tagExists;

        } catch (SQLException e) {

            String msg = "Failed to check the existence of the tag " + tagName +
                    " on resource " + resourceID + " by user " + userID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

    }

    public static void removeTag(String path, String tag)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResourceDAO resourceDAO = new ResourceDAO();
        String resourceID = ResourceDAO.getResourceID(path);
        PreparedStatement ps = null;
        try {
            String sql = "DELETE FROM REG_TAG WHERE REG_RID=? AND REG_TAG_NAME=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setString(2, tag);

            ps.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to remove tag " + tag + " from the resource " +
                    path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public static void removeTag(String path, String tag, String userName)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResourceDAO resourceDAO = new ResourceDAO();
        String resourceID = ResourceDAO.getResourceID(path);
        PreparedStatement s = null;
        try {

            String sql = "DELETE FROM REG_TAG WHERE REG_RID=? AND REG_TAG_NAME=?";
            if (!userName.equals("*")) {
                sql = sql + "AND REG_USER_ID=?";
            }

            s = conn.prepareStatement(sql);
            s.setString(1, resourceID);
            s.setString(2, tag);
            if (!userName.equals("*")) {
                s.setString(3, userName);
            }

            s.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to remove tag " + tag + " on resource " + path +
                    " done by user " + userName + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public static String[] getPathsWithTag(String tag, Connection conn) throws SQLException, RegistryException {

        String sql = "SELECT R.REG_PATH FROM REG_RESOURCE R, REG_TAG T WHERE T.REG_TAG_NAME=? AND T.REG_RID=R.REG_RID";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, tag);

        ResultSet result = s.executeQuery();
        List pathList = new ArrayList();
        while (result.next()) {
            pathList.add(result.getString(DatabaseConstants.PATH_FIELD));
        }

        try {
            if (result != null) {
                result.close();
            }
            if (s != null) {
                s.close();
            }
        }
        catch (SQLException ex) {
            String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
            log.error(msg, ex);
            throw new RegistryException(msg, ex);
        }

        return (String[]) pathList.toArray(new String[pathList.size()]);
    }

    public static String[] getTags(String path) throws RegistryException {

        Connection conn = Transaction.getConnection();
        ResultSet result = null;
        PreparedStatement s = null;
        try {
            String sql = "SELECT T.REG_TAG_NAME FROM REG_TAG T, REG_RESOURCE R WHERE R.REG_PATH=? AND R.REG_RID=T.REG_RID";

            s = conn.prepareStatement(sql);
            s.setString(1, path);

            result = s.executeQuery();
            List tagList = new ArrayList();
            while (result.next()) {
                tagList.add(result.getString(DatabaseConstants.TAG_NAME_FIELD));
            }

            return (String[]) tagList.toArray(new String[tagList.size()]);

        } catch (SQLException e) {

            String msg = "Failed to get tags associated with the resource path " +
                    path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public static List getPathsWithAnyTag(String[] tags) throws RegistryException {

        if (tags == null || tags.length == 0) return null;

        Connection conn = Transaction.getConnection();

        StringBuffer sqlBuf = new StringBuffer();
        sqlBuf.append("SELECT DISTINCT R.REG_PATH FROM REG_RESOURCE R, REG_TAG T WHERE R.REG_RID=T.REG_RID");
        sqlBuf.append(" AND (");
        for (int i = 0; i < tags.length; i++) {
            if (i > 0) sqlBuf.append(" OR ");
            sqlBuf.append("lower(T.REG_TAG_NAME)=lower(?)");
        }
        sqlBuf.append(")");

        List resourcePaths = new ArrayList();

        ResultSet results = null;
        PreparedStatement s = null;
        try {
            s = conn.prepareStatement(sqlBuf.toString());

            for (int i = 0; i < tags.length; i++) {
                s.setString(i + 1, tags[i]);
            }

            results = s.executeQuery();
            while (results.next()) {
                resourcePaths.add(results.getString(DatabaseConstants.PATH_FIELD));
            }

        } catch (SQLException e) {

            String msg = "Failed to resource paths with any of the tags " +
                    Arrays.toString(tags) + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return resourcePaths;
    }

    public static long getTagCount(String path, String tag) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql =
                "SELECT COUNT(T.REG_TAG_NAME) FROM REG_RESOURCE R, REG_TAG T " +
                        "WHERE R.REG_PATH=? AND lower(T.REG_TAG_NAME)=lower(?) AND " +
                        "T.REG_RID=R.REG_RID GROUP BY R.REG_PATH";

        ResultSet result = null;
        PreparedStatement s = null;
        try {
            s = conn.prepareStatement(sql);
            s.setString(1, path);
            s.setString(2, tag);

            result = s.executeQuery();

            long tagCount = 0;
            if (result.next()) {
                tagCount = result.getLong(1);
            }

            return tagCount;

        } catch (SQLException e) {

            String msg = "Failed to get tag count of tag " + tag +
                    " on resource " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public static List getTaggedPaths(String tag, Connection conn) throws SQLException, RegistryException {

        String sql =
                "SELECT A.PATH, COUNT(TN.TAG_NAME) FROM ARTIFACTS A, TAGS TN " +
                        "WHERE lower(TN.TAG_NAME)=lower(?) AND TN.AID=A.AID GROUP BY A.PATH";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, tag);

        ResultSet result = s.executeQuery();
        List pathList = new ArrayList();
        while (result.next()) {
            TaggedResourcePath taggedResourcePath = new TaggedResourcePath();
            taggedResourcePath.setResourcePath(result.getString(DatabaseConstants.PATH_FIELD));
            taggedResourcePath.setTagCount(result.getLong(2));
            pathList.add(taggedResourcePath);
        }

        try {
            if (result != null) {
                result.close();
            }
            if (s != null) {
                s.close();
            }
        }
        catch (SQLException ex) {
            String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
            log.error(msg, ex);
            throw new RegistryException(msg, ex);
        }
        return pathList;
    }

    public static Tag[] getTagsWithCount(String path) throws RegistryException {

        Connection conn = Transaction.getConnection();

        List tagList = new ArrayList();
        ResultSet result = null;
        PreparedStatement s = null;
        try {
            String sql = "SELECT T.REG_TAG_NAME, COUNT(R.REG_RID) FROM REG_TAG T, REG_RESOURCE R " +
                    "WHERE R.REG_PATH=? AND R.REG_RID=T.REG_RID GROUP BY T.REG_TAG_NAME";

            s = conn.prepareStatement(sql);
            s.setString(1, path);

            result = s.executeQuery();
            while (result.next()) {
                Tag tag = new Tag();
                tag.setTagName(result.getString(DatabaseConstants.TAG_NAME_FIELD));
                tag.setTagCount(result.getLong(2));
                tagList.add(tag);
            }

        } catch (SQLException e) {

            String msg = "Failed to get tags and tag counts of the resource " +
                    path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return (Tag[]) tagList.toArray(new Tag[tagList.size()]);
    }

    public static TaggingDO getTagging(String path, String tagName, String userName)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResourceDAO resourceDAO = new ResourceDAO();
        String resourceID = ResourceDAO.getResourceID(path);

        ResultSet result = null;
        PreparedStatement s = null;
        try {
            TaggingDO taggingDO = null;

            String sql = "SELECT T.REG_TAGGED_TIME FROM REG_TAG T, REG_RESOURCE R " +
                    "WHERE R.REG_RID=T.REG_RID AND T.REG_RID=? AND T.REG_USER_ID=? AND T.REG_TAG_NAME=?";

            s = conn.prepareStatement(sql);
            s.setString(1, resourceID);
            s.setString(2, userName);
            s.setString(3, tagName);

            result = s.executeQuery();
            if (result.next()) {

                java.util.Date taggedTime = new java.util.Date(
                        result.getTimestamp(DatabaseConstants.TAGGED_TIME_FIELD).getTime());

                taggingDO = new TaggingDO();
                taggingDO.setResourceID(resourceID);
                taggingDO.setResourcePath(path);
                taggingDO.setTagName(tagName);
                taggingDO.setTaggedTime(taggedTime);
                taggingDO.setTaggedUserName(userName);
            }

            return taggingDO;

        } catch (SQLException e) {

            String msg = "Failed to get tagging information for tag " + tagName +
                    " on resource " + path + " by user " + userName + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public static TaggingDO getTagging(long taggingID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet result = null;
        PreparedStatement s = null;
        try {
            String sql = "SELECT R.REG_RID, R.REG_PATH, T.REG_USER_ID, T.REG_TAG_NAME, T.REG_TAGGED_TIME " +
                    "FROM REG_TAG T, REG_RESOURCE R WHERE R.REG_RID=T.REG_RID AND T.REG_TAG_ID=?";

            s = conn.prepareStatement(sql);
            s.setLong(1, taggingID);

            TaggingDO taggingDO = null;
            result = s.executeQuery();
            if (result.next()) {

                java.util.Date taggedTime = new java.util.Date(
                        result.getTimestamp(DatabaseConstants.TAGGED_TIME_FIELD).getTime());

                taggingDO = new TaggingDO();
                taggingDO.setResourceID(result.getString(DatabaseConstants.AID_FIELD));
                taggingDO.setResourcePath(result.getString(DatabaseConstants.PATH_FIELD));
                taggingDO.setTagName(result.getString(DatabaseConstants.TAG_NAME_FIELD));
                taggingDO.setTaggedUserName(result.getString(DatabaseConstants.USER_ID_FIELD));
                taggingDO.setTaggedTime(taggedTime);
            }

            return taggingDO;

        } catch (SQLException e) {

            String msg = "Failed to get tagging information for tag ID " +
                    taggingID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }
}
