/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.utils.creators;

import javax.sql.DataSource;
import java.sql.SQLException;
import java.sql.Statement;

public class HSQLDatabaseCreator extends DatabaseCreator {

    private static final String clusterLockTable =
            "CREATE TABLE REG_CLUSTER_LOCK (" +
                    "REG_LOCK_NAME VARCHAR (20)," +
                    "REG_LOCK_STATUS VARCHAR (20)," +
                    "REG_LOCKED_TIME TIMESTAMP," +
                    "PRIMARY KEY (REG_LOCK_NAME))";

    private static final String resourceTable =
            "CREATE TABLE REG_RESOURCE (" +
                    "REG_RID VARCHAR (50)," +
                    "REG_PATH VARCHAR (2000) NOT NULL," +
                    "REG_MEDIA_TYPE VARCHAR (500)," +
                    "REG_COLLECTION INTEGER NOT NULL," +
                    "REG_CREATOR VARCHAR (500)," +
                    "REG_CREATED_TIME TIMESTAMP," +
                    "REG_LAST_UPDATOR VARCHAR (500)," +
                    "REG_LAST_UPDATED_TIME TIMESTAMP," +
                    "REG_DESCRIPTION VARCHAR (10000)," +
                    "REG_CONTENT_ID VARCHAR (50)," +
                    "REG_EQUIVALENT_VERSION INTEGER NOT NULL," +
                    "REG_ASSOCIATED_SNAPSHOT_ID INTEGER NOT NULL," +
                    "PRIMARY KEY (REG_RID)," +
                    "FOREIGN KEY (REG_CONTENT_ID) REFERENCES REG_CONTENT (REG_CONTENT_ID)," +
                    "UNIQUE(REG_PATH));";

    private static final String contentTable =
            "CREATE TABLE REG_CONTENT (" +
                    "REG_CONTENT_ID VARCHAR (50)," +
                    "REG_CONTENT_DATA BINARY," +
                    "PRIMARY KEY (REG_CONTENT_ID));";

    // Table to store parent child relationships in current versions of resources. Collection
    // which is referred in PARENT_RID is not allowed to delete until the relationship is
    // explicitly removed from this table. This is to ensure that collections are not deleted
    // till all its children are removed. If a child resource (which has no children) is deleted
    // its parent child relationship is automatically deleted.

    private static final String dependencyTable =
            "CREATE TABLE REG_DEPENDENCY (" +
                    "REG_DEPENDENCY_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
                    "REG_PARENT_RID VARCHAR (50) NOT NULL," +
                    "REG_CHILD_RID VARCHAR (50) NOT NULL," +
                    "PRIMARY KEY (REG_DEPENDENCY_ID)," +
                    "UNIQUE (REG_PARENT_RID, REG_CHILD_RID)," +
                    "FOREIGN KEY (REG_PARENT_RID) REFERENCES REG_RESOURCE (REG_RID) ON DELETE CASCADE," +
                    "FOREIGN KEY (REG_CHILD_RID) REFERENCES REG_RESOURCE (REG_RID) ON DELETE CASCADE);";

    // We can't make (RID, NAME, PROPERTY_VALUE) a unique tuple as resource may have a property with
    // repeating values. e.g.p1 -> v1, v2, v3, v3, v4
    private static final String propertyTable =
            "CREATE TABLE REG_PROPERTY (" +
                    "REG_PROPERTY_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
                    "REG_RID VARCHAR (50) NOT NULL," +
                    "REG_NAME VARCHAR (100) NOT NULL," +
                    "REG_PROPERTY_VALUE VARCHAR (500)," +
                    "PRIMARY KEY (REG_PROPERTY_ID))";

    private static final String associationTable =
            "CREATE TABLE REG_ASSOCIATION (" +
                    "REG_ASSOCIATION_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
                    "REG_SOURCEPATH VARCHAR (2000) NOT NULL," +
                    "REG_TARGETPATH VARCHAR (2000) NOT NULL," +
                    "REG_ASSOCIATION_TYPE VARCHAR (2000) NOT NULL," +
                    "PRIMARY KEY (REG_ASSOCIATION_ID)," +
                    "UNIQUE (REG_SOURCEPATH, REG_TARGETPATH, REG_ASSOCIATION_TYPE))";

    // Note that comments, tags and ratings are not versioned. Therefore, when a resource is deleted
    // its comments, etc should not be deleted. When a deleted resource is restored, its
    // comments, etc are reassociated with it.

    private static final String tagTable =
            "CREATE TABLE REG_TAG (" +
                    "REG_TAG_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
                    "REG_TAG_NAME VARCHAR (500) NOT NULL," +
                    "REG_RID VARCHAR (50) NOT NULL," +
                    "REG_USER_ID VARCHAR (20) NOT NULL," +
                    "REG_TAGGED_TIME TIMESTAMP NOT NULL," +
                    "PRIMARY KEY (REG_TAG_ID)," +
                    "UNIQUE (REG_TAG_NAME, REG_RID, REG_USER_ID));";

    private static final String commentTable =
            "CREATE TABLE REG_COMMENT (" +
                    "REG_COMMENT_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
                    "REG_RID VARCHAR (50) NOT NULL," +
                    "REG_USER_ID VARCHAR (20) NOT NULL," +
                    "REG_COMMENT_TEXT VARCHAR (500) NOT NULL," +
                    "REG_COMMENTED_TIME TIMESTAMP NOT NULL," +
                    "PRIMARY KEY (REG_COMMENT_ID));";

    private static final String ratingTable =
            "CREATE TABLE REG_RATING (" +
                    "REG_RATING_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
                    "REG_RID VARCHAR (50) NOT NULL," +
                    "REG_USER_ID VARCHAR (20) NOT NULL," +
                    "REG_RATING INTEGER NOT NULL," +
                    "REG_RATED_TIME TIMESTAMP NOT NULL," +
                    "PRIMARY KEY (REG_RATING_ID));";

    private static final String logTable =
            "CREATE TABLE REG_LOG (" +
                    "REG_LOG_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
                    "REG_PATH VARCHAR (2000)," +
                    "REG_USER_ID VARCHAR (20) NOT NULL," +
                    "REG_LOGGED_TIME TIMESTAMP NOT NULL," +
                    "REG_ACTION INTEGER NOT NULL," +
                    "REG_ACTION_DATA VARCHAR (500)," +
                    "PRIMARY KEY (REG_LOG_ID));";

    //////////////////////////////////////////////////////////////////////////////////
    // Tables related to old versions of resources and metadata                     //
    //////////////////////////////////////////////////////////////////////////////////

    private static final String resourceVersionTable =
            "CREATE TABLE REG_RESOURCE_VERSION (" +
                    "REG_RESOURCE_VERSION_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
                    "REG_RID VARCHAR (50)," +
                    "REG_VERSION INTEGER NOT NULL," +
                    "REG_PATH VARCHAR (2000) NOT NULL," +
                    "REG_MEDIA_TYPE VARCHAR (500)," +
                    "REG_COLLECTION INTEGER NOT NULL," +
                    "REG_CREATOR VARCHAR (500)," +
                    "REG_CREATED_TIME TIMESTAMP," +
                    "REG_LAST_UPDATOR VARCHAR (500)," +
                    "REG_LAST_UPDATED_TIME TIMESTAMP," +
                    "REG_DESCRIPTION VARCHAR (10000)," +
                    "REG_CONTENT_ID VARCHAR (50)," +
                    "REG_ASSOCIATED_SNAPSHOT_ID INTEGER NOT NULL," +
                    "FOREIGN KEY (REG_CONTENT_ID) REFERENCES REG_CONTENT_VERSION (REG_CONTENT_VERSION_ID)," +
                    "PRIMARY KEY (REG_RESOURCE_VERSION_ID)," +
                    "UNIQUE (REG_RID, REG_VERSION))";

    private static final String dependencyVersionTable =
            "CREATE TABLE REG_DEPENDENCY_VERSION (" +
                    "REG_DEPENDENCY_VERSION_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
                    "REG_PARENT_RID VARCHAR (50) NOT NULL," +
                    "REG_PARENT_VERSION INTEGER NOT NULL," +
                    "REG_CHILD_RID VARCHAR (50) NOT NULL," +
                    "PRIMARY KEY (REG_DEPENDENCY_VERSION_ID)," +
                    "UNIQUE (REG_PARENT_RID, REG_PARENT_VERSION, REG_CHILD_RID))";

    private static final String contentVersionTable =
            "CREATE TABLE REG_CONTENT_VERSION (" +
                    "REG_CONTENT_VERSION_ID VARCHAR (50)," +
                    "REG_CONTENT_DATA BINARY," +
                    "PRIMARY KEY (REG_CONTENT_VERSION_ID));";

    private static final String propertyVersionTable =
            "CREATE TABLE REG_PROPERTY_VERSION (" +
                    "REG_PROPERTY_VERSION_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
                    "REG_RID VARCHAR (50) NOT NULL," +
                    "REG_VERSION INTEGER NOT NULL," +
                    "REG_NAME VARCHAR (100) NOT NULL," +
                    "REG_PROPERTY_VALUE VARCHAR (500)," +
                    "PRIMARY KEY (REG_PROPERTY_VERSION_ID)," +
                    "FOREIGN KEY (REG_RID, REG_VERSION) REFERENCES REG_RESOURCE_VERSION (REG_RID, REG_VERSION));";

    //////////////////////////////////////////////////////////////////////////////////
    // Tables related to snapshot management                                        //
    //////////////////////////////////////////////////////////////////////////////////

    // UUID is used for snapshot ID, so that snapshots can be achieved and reloaded without
    // colliding with existing snapshots.

    private static final String snapshotTable =
            "CREATE TABLE REG_SNAPSHOT (" +
                    "REG_SNAPSHOT_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
                    "REG_ROOT_ID VARCHAR (50) NOT NULL," +
                    "PRIMARY KEY (REG_SNAPSHOT_ID)," +
                    "UNIQUE (REG_SNAPSHOT_ID, REG_ROOT_ID));";

    private static final String snapshotRootIDIndex =
            "CREATE INDEX REG_INDEX_SNAPSHOT_ROOT_ID ON REG_SNAPSHOT (REG_ROOT_ID)";

    private static final String snapshotResourceVersionTable =
            "CREATE TABLE REG_SNAPSHOT_RESOURCE_VERSION (" +
                    "REG_SRV_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
                    "REG_SNAPSHOT_ID INTEGER NOT NULL," +
                    "REG_RID VARCHAR (50) NOT NULL," +
                    "REG_VERSION INTEGER NOT NULL," +
                    "PRIMARY KEY (REG_SRV_ID)," +
                    "UNIQUE (REG_SNAPSHOT_ID, REG_RID, REG_VERSION));";

    ////////////////////////////////////////////////////////////////////////////////////
    // User manager tables                                                            //
    ////////////////////////////////////////////////////////////////////////////////////

    public static String userTable = "CREATE TABLE UM_USERS ("
            + "ID INTEGER GENERATED BY DEFAULT AS IDENTITY," + "USER_NAME VARCHAR(255) NOT NULL,"
            + "USER_PASSWORD VARCHAR(255) NOT NULL, " + "PRIMARY KEY (ID), " + "UNIQUE(USER_NAME))";

    public static String userAttributeTable = "CREATE TABLE UM_USER_ATTRIBUTES ("
            + "ID INTEGER GENERATED BY DEFAULT AS IDENTITY," + "ATTR_NAME VARCHAR(255) NOT NULL,"
            + "ATTR_VALUE VARCHAR(255), " + "USER_ID INTEGER, "
            + "FOREIGN KEY (USER_ID) REFERENCES UM_USERS(ID) ON DELETE CASCADE,"
            + "PRIMARY KEY (ID))";

    public static String roleTable = "CREATE TABLE UM_ROLES ("
            + "ID INTEGER GENERATED BY DEFAULT AS IDENTITY, " + "ROLE_NAME VARCHAR(255) NOT NULL, "
            + "PRIMARY KEY (ID), " + "UNIQUE(ROLE_NAME))";

    public static String roleAttributeTable = "CREATE TABLE UM_ROLE_ATTRIBUTES ("
            + "ID INTEGER GENERATED BY DEFAULT AS IDENTITY, " + "ATTR_NAME VARCHAR(255) NOT NULL, "
            + "ATTR_VALUE VARCHAR(255), " + "ROLE_ID INTEGER, "
            + "FOREIGN KEY (ROLE_ID) REFERENCES UM_ROLES(ID) ON DELETE CASCADE,"
            + "PRIMARY KEY (ID))";

    public static String permissionTable = "CREATE TABLE UM_PERMISSIONS ("
            + "ID INTEGER GENERATED BY DEFAULT AS IDENTITY, "
            + "RESOURCE_ID VARCHAR(255) NOT NULL, " + "ACTION VARCHAR(255) NOT NULL, "
            + "PRIMARY KEY (ID))";

    public static String permissionResourceIDActionIndex =
            "CREATE INDEX INDEX_UM_PERMISSIONS_RESOURCE_ID_ACTION " +
                    "ON UM_PERMISSIONS (RESOURCE_ID, ACTION)";

    public static String rolePermissionTable = "CREATE TABLE UM_ROLE_PERMISSIONS ( "
            + "ID INTEGER GENERATED BY DEFAULT AS IDENTITY, " + "PERMISSION_ID INTEGER NOT NULL, "
            + "ROLE_ID INTEGER NOT NULL, " + "IS_ALLOWED SMALLINT NOT NULL, "
            + "UNIQUE (PERMISSION_ID, ROLE_ID), "
            + "FOREIGN KEY (PERMISSION_ID) REFERENCES UM_PERMISSIONS(ID) ON DELETE  CASCADE, "
            + "FOREIGN KEY (ROLE_ID) REFERENCES UM_ROLES(ID) ON DELETE CASCADE, "
            + "PRIMARY KEY (ID))";

    public static String userPermissionTable = "CREATE TABLE UM_USER_PERMISSIONS ("
            + "ID INTEGER GENERATED BY DEFAULT AS IDENTITY, " + "PERMISSION_ID INTEGER NOT NULL, "
            + "USER_ID INTEGER NOT NULL, " + "IS_ALLOWED SMALLINT NOT NULL, "
            + "UNIQUE (PERMISSION_ID, USER_ID), "
            + "FOREIGN KEY (PERMISSION_ID) REFERENCES UM_PERMISSIONS(ID) ON DELETE CASCADE, "
            + "FOREIGN KEY (USER_ID) REFERENCES UM_USERS(ID) ON DELETE CASCADE, "
            + "PRIMARY KEY (ID))";

    public static String userRolesTable = "CREATE TABLE UM_USER_ROLES ("
            + "ID INTEGER GENERATED BY DEFAULT AS IDENTITY, " + "ROLE_ID INTEGER NOT NULL, "
            + "USER_ID INTEGER NOT NULL, " + "UNIQUE (USER_ID, ROLE_ID), "
            + "FOREIGN KEY (ROLE_ID) REFERENCES UM_ROLES(ID) ON DELETE CASCADE, "
            + "FOREIGN KEY (USER_ID) REFERENCES UM_USERS(ID) ON DELETE CASCADE, "
            + "PRIMARY KEY (ID))";

    public HSQLDatabaseCreator(DataSource dataSource) {
        super(dataSource);
    }

    public String getClusterLockTable() {
        return clusterLockTable;
    }

    public String getResourceTable() {
        return resourceTable;
    }

    public String getContentTable() {
        return contentTable;
    }

    public String getDependencyTable() {
        return dependencyTable;
    }

    public String getPropertyTable() {
        return propertyTable;
    }

    public String getAssociationTable() {
        return associationTable;
    }

    public String getTagTable() {
        return tagTable;
    }

    public String getCommentTable() {
        return commentTable;
    }

    public String getRatingTable() {
        return ratingTable;
    }

    public String getLogTable() {
        return logTable;
    }

    public String getResourceVersionTable() {
        return resourceVersionTable;
    }

    public String getDependencyVersionTable() {
        return dependencyVersionTable;
    }

    public String getContentVersionTable() {
        return contentVersionTable;
    }

    public String getPropertyVersionTable() {
        return propertyVersionTable;
    }

    public String getSnapshotTable() {
        return snapshotTable;
    }

    public String getSnapshotResourceVersionTable() {
        return snapshotResourceVersionTable;
    }

    public String getUserTable() {
        return userTable;
    }

    public String getUserAttributeTable() {
        return userAttributeTable;
    }

    public String getRoleTable() {
        return roleTable;
    }

    public String getRoleAttributeTable() {
        return roleAttributeTable;
    }

    public String getPermissionTable() {
        return permissionTable;
    }

    public String getRolePermissionTable() {
        return rolePermissionTable;
    }

    public String getUserPermissionTable() {
        return userPermissionTable;
    }

    public String getUserRolesTable() {
        return userRolesTable;
    }

    public String getSnapshotRootIDIndex() {
        return snapshotRootIDIndex;
    }

    public void executeResourceStoreDatabaseSpecificStatements(Statement s) throws SQLException {
    }

    public void executeUserStoreDatabaseSpecificStatements(Statement s) throws SQLException {
        s.executeUpdate(permissionResourceIDActionIndex);
    }
}
