/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.utils.creators;

import javax.sql.DataSource;
import java.sql.SQLException;
import java.sql.Statement;

public class OracleDatabaseCreator extends DatabaseCreator {

    private static final String clusterLockTable =
            "CREATE TABLE REG_CLUSTER_LOCK (" +
                    "REG_LOCK_NAME VARCHAR (20)," +
                    "REG_LOCK_STATUS VARCHAR (20)," +
                    "REG_LOCKED_TIME TIMESTAMP," +
                    "PRIMARY KEY (REG_LOCK_NAME))";

    private static final String resourceTable =
            "CREATE TABLE REG_RESOURCE (" +
                    "REG_RID VARCHAR2 (50)," +
                    "REG_PATH VARCHAR2 (2000) NOT NULL," +
                    "REG_MEDIA_TYPE VARCHAR2 (500)," +
                    "REG_COLLECTION INTEGER NOT NULL," +
                    "REG_CREATOR VARCHAR2 (500)," +
                    "REG_CREATED_TIME TIMESTAMP," +
                    "REG_LAST_UPDATOR VARCHAR2 (500)," +
                    "REG_LAST_UPDATED_TIME TIMESTAMP," +
                    "REG_DESCRIPTION VARCHAR2 (4000)," +
                    "REG_CONTENT_ID VARCHAR2 (50)," +
                    "REG_EQUIVALENT_VERSION INTEGER NOT NULL," +
                    "REG_ASSOCIATED_SNAPSHOT_ID INTEGER NOT NULL," +
                    "PRIMARY KEY (REG_RID)," +
                    "FOREIGN KEY (REG_CONTENT_ID) REFERENCES REG_CONTENT (REG_CONTENT_ID))";

    private static final String resourcePathIndex =
            "CREATE INDEX REG_RESOURCE_PATH ON REG_RESOURCE (REG_PATH)";

    private static final String contentTable =
            "CREATE TABLE REG_CONTENT (" +
                    "REG_CONTENT_ID VARCHAR2 (50)," +
                    "REG_CONTENT_DATA BLOB," +
                    "PRIMARY KEY (REG_CONTENT_ID))";

    // Table to store parent child relationships in current versions of resources. Collection
    // which is referred in PARENT_RID is not allowed to delete until the relationship is
    // explicitly removed from this table. This is to ensure that collections are not deleted
    // till all its children are removed. If a child resource (which has no children) is deleted
    // its parent child relationship is automatically deleted.

    private static final String dependencyTable =
            "CREATE TABLE REG_DEPENDENCY (" +
                    "REG_DEPENDENCY_ID INTEGER," +
                    "REG_PARENT_RID VARCHAR2 (50) NOT NULL," +
                    "REG_CHILD_RID VARCHAR2 (50) NOT NULL," +
                    "PRIMARY KEY (REG_DEPENDENCY_ID)," +
                    "UNIQUE (REG_PARENT_RID, REG_CHILD_RID)," +
                    "FOREIGN KEY (REG_PARENT_RID) REFERENCES REG_RESOURCE (REG_RID) ON DELETE CASCADE," +
                    "FOREIGN KEY (REG_CHILD_RID) REFERENCES REG_RESOURCE (REG_RID) ON DELETE CASCADE)";

    private static final String dependencySequence =
            "CREATE SEQUENCE REG_DEPENDENCY_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    private static final String dependencyIDIncrementTrigger =
            "CREATE OR REPLACE TRIGGER REG_DEPENDENCY_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON REG_DEPENDENCY " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.REG_PARENT_RID IS NOT NULL AND :NEW.REG_CHILD_RID IS NOT NULL THEN " +
                    "SELECT REG_DEPENDENCY_SEQUENCE.nextval INTO :NEW.REG_DEPENDENCY_ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    // We can't make (RID, NAME, PROPERTY_VALUE) a unique tuple as resource may have a property with
    // repeating values. e.g.p1 -> v1, v2, v3, v3, v4
    private static final String propertyTable =
            "CREATE TABLE REG_PROPERTY (" +
                    "REG_PROPERTY_ID INTEGER," +
                    "REG_RID VARCHAR2 (50) NOT NULL," +
                    "REG_NAME VARCHAR2 (100) NOT NULL," +
                    "REG_PROPERTY_VALUE VARCHAR2 (500)," +
                    "PRIMARY KEY (REG_PROPERTY_ID))";

    private static final String propertySequence =
            "CREATE SEQUENCE REG_PROPERTY_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    private static final String propertyIDTrigger =
            "CREATE OR REPLACE TRIGGER REG_PROPERTY_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON REG_PROPERTY " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.REG_RID IS NOT NULL AND :NEW.REG_NAME IS NOT NULL THEN " +
                    "SELECT REG_PROPERTY_SEQUENCE.nextval INTO :NEW.REG_PROPERTY_ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    private static final String associationTable =
            "CREATE TABLE REG_ASSOCIATION (" +
                    "REG_ASSOCIATION_ID INTEGER," +
                    "REG_SOURCEPATH VARCHAR2 (2000) NOT NULL," +
                    "REG_TARGETPATH VARCHAR2 (2000) NOT NULL," +
                    "REG_ASSOCIATION_TYPE VARCHAR2 (2000) NOT NULL," +
                    "PRIMARY KEY (REG_ASSOCIATION_ID))";

    private static final String associationSequence =
            "CREATE SEQUENCE REG_ASSOCIATION_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    private static final String associationIDTrigger =
            "CREATE OR REPLACE TRIGGER REG_ASSOCIATION_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON REG_ASSOCIATION " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.REG_SOURCEPATH IS NOT NULL AND :NEW.REG_TARGETPATH IS NOT NULL " +
                    "AND :NEW.REG_ASSOCIATION_TYPE IS NOT NULL THEN " +
                    "SELECT REG_ASSOCIATION_SEQUENCE.nextval INTO :NEW.REG_ASSOCIATION_ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    private static final String sourcePathIndex =
            "CREATE INDEX REG_ASSOCIATION_SOURCEPATH ON REG_ASSOCIATION (REG_SOURCEPATH)";

    private static final String targetPathIndex =
            "CREATE INDEX REG_ASSOCIATION_TARGETPATH ON REG_ASSOCIATION (REG_TARGETPATH)";

    // Note that comments, tags and ratings are not versioned. Therefore, when a resource is deleted
    // its comments, etc should not be deleted. When a deleted resource is restored, its
    // comments, etc are reassociated with it.

    private static final String tagTable =
            "CREATE TABLE REG_TAG (" +
                    "REG_TAG_ID INTEGER," +
                    "REG_TAG_NAME VARCHAR2 (500) NOT NULL," +
                    "REG_RID VARCHAR2 (50) NOT NULL," +
                    "REG_USER_ID VARCHAR2 (20) NOT NULL," +
                    "REG_TAGGED_TIME TIMESTAMP NOT NULL," +
                    "PRIMARY KEY (REG_TAG_ID)," +
                    "UNIQUE (REG_TAG_NAME, REG_RID, REG_USER_ID))";

    private static final String tagIDSequence =
            "CREATE SEQUENCE REG_TAG_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    private static final String tagIDTrigger =
            "CREATE OR REPLACE TRIGGER REG_TAG_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON REG_TAG " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.REG_TAG_NAME IS NOT NULL AND :NEW.REG_RID IS NOT NULL AND :NEW.REG_USER_ID IS NOT NULL " +
                    "AND :NEW.REG_TAGGED_TIME IS NOT NULL THEN " +
                    "SELECT REG_TAG_SEQUENCE.nextval INTO :NEW.REG_TAG_ID FROM dual; " +
                    " END IF; " +
                    "END; ";

    private static final String commentTable =
            "CREATE TABLE REG_COMMENT (" +
                    "REG_COMMENT_ID INTEGER," +
                    "REG_RID VARCHAR2 (50) NOT NULL," +
                    "REG_USER_ID VARCHAR2 (20) NOT NULL," +
                    "REG_COMMENT_TEXT VARCHAR2 (500) NOT NULL," +
                    "REG_COMMENTED_TIME TIMESTAMP NOT NULL," +
                    "PRIMARY KEY (REG_COMMENT_ID))";

    private static final String commentIDSequence =
            "CREATE SEQUENCE REG_COMMENT_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    private static final String commentIDTrigger =
            "CREATE OR REPLACE TRIGGER REG_COMMENT_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON REG_COMMENT " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.REG_RID IS NOT NULL AND :NEW.REG_USER_ID IS NOT NULL AND :NEW.REG_COMMENT_TEXT IS NOT NULL " +
                    "AND :NEW.REG_COMMENTED_TIME IS NOT NULL THEN " +
                    "SELECT REG_COMMENT_SEQUENCE.nextval INTO :NEW.REG_COMMENT_ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    private static final String ratingTable =
            "CREATE TABLE REG_RATING (" +
                    "REG_RATING_ID INTEGER," +
                    "REG_RID VARCHAR2 (50) NOT NULL," +
                    "REG_USER_ID VARCHAR2 (20) NOT NULL," +
                    "REG_RATING INTEGER NOT NULL," +
                    "REG_RATED_TIME TIMESTAMP NOT NULL," +
                    "PRIMARY KEY (REG_RATING_ID))";

    private static final String ratingIDSequence =
            "CREATE SEQUENCE REG_RATING_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    private static final String ratingIDTrigger =
            "CREATE OR REPLACE TRIGGER REG_RATING_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON REG_RATING " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.REG_RID IS NOT NULL AND :NEW.REG_USER_ID IS NOT NULL AND :NEW.REG_RATING IS NOT NULL " +
                    "AND :NEW.REG_RATED_TIME IS NOT NULL THEN " +
                    "SELECT REG_RATING_SEQUENCE.nextval INTO :NEW.REG_RATING_ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    private static final String logTable =
            "CREATE TABLE REG_LOG (" +
                    "REG_LOG_ID INTEGER," +
                    "REG_PATH VARCHAR2 (2000)," +
                    "REG_USER_ID VARCHAR2 (20) NOT NULL," +
                    "REG_LOGGED_TIME TIMESTAMP NOT NULL," +
                    "REG_ACTION INTEGER NOT NULL," +
                    "REG_ACTION_DATA VARCHAR2 (500)," +
                    "PRIMARY KEY (REG_LOG_ID))";

    private static final String logIDSequence =
            "CREATE SEQUENCE REG_LOG_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    private static final String logIDTrigger =
            "CREATE OR REPLACE TRIGGER REG_LOG_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON REG_LOG " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.REG_USER_ID IS NOT NULL AND :NEW.REG_LOGGED_TIME IS NOT NULL " +
                    "AND :NEW.REG_ACTION IS NOT NULL THEN " +
                    "SELECT REG_LOG_SEQUENCE.nextval INTO :NEW.REG_LOG_ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    //////////////////////////////////////////////////////////////////////////////////
    // Tables related to old versions of resources and metadata                     //
    //////////////////////////////////////////////////////////////////////////////////

    private static final String resourceVersionTable =
            "CREATE TABLE REG_RESOURCE_VERSION (" +
                    "REG_RESOURCE_VERSION_ID INTEGER," +
                    "REG_RID VARCHAR2 (50) NOT NULL," +
                    "REG_VERSION INTEGER NOT NULL," +
                    "REG_PATH VARCHAR2 (2000) NOT NULL," +
                    "REG_MEDIA_TYPE VARCHAR2 (500)," +
                    "REG_COLLECTION INTEGER NOT NULL," +
                    "REG_CREATOR VARCHAR2 (500)," +
                    "REG_CREATED_TIME TIMESTAMP," +
                    "REG_LAST_UPDATOR VARCHAR2 (500)," +
                    "REG_LAST_UPDATED_TIME TIMESTAMP," +
                    "REG_DESCRIPTION VARCHAR2 (4000)," +
                    "REG_CONTENT_ID VARCHAR2 (50)," +
                    "REG_ASSOCIATED_SNAPSHOT_ID INTEGER NOT NULL," +
                    "FOREIGN KEY (REG_CONTENT_ID) REFERENCES REG_CONTENT_VERSION (REG_CONTENT_VERSION_ID)," +
                    "PRIMARY KEY (REG_RESOURCE_VERSION_ID)," +
                    "UNIQUE(REG_RID, REG_VERSION))";

    private static final String resourceVersionIDSequence =
            "CREATE SEQUENCE REG_RESOURCE_VERSION_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    private static final String resourceVersionIDTrigger =
            "CREATE OR REPLACE TRIGGER REG_RESOURCE_VERSION_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON REG_RESOURCE_VERSION " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.REG_RID IS NOT NULL AND :NEW.REG_VERSION IS NOT NULL AND :NEW.REG_PATH IS NOT NULL " +
                    "AND :NEW.REG_COLLECTION IS NOT NULL AND :NEW.REG_ASSOCIATED_SNAPSHOT_ID IS NOT NULL THEN " +
                    "SELECT REG_RESOURCE_VERSION_SEQUENCE.nextval INTO :NEW.REG_RESOURCE_VERSION_ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    private static final String dependencyVersionTable =
            "CREATE TABLE REG_DEPENDENCY_VERSION (" +
                    "REG_DEPENDENCY_VERSION_ID INTEGER," +
                    "REG_PARENT_RID VARCHAR2 (50) NOT NULL," +
                    "REG_PARENT_VERSION INTEGER NOT NULL," +
                    "REG_CHILD_RID VARCHAR2 (50) NOT NULL," +
                    "PRIMARY KEY (REG_DEPENDENCY_VERSION_ID)," +
                    "UNIQUE (REG_PARENT_RID, REG_PARENT_VERSION, REG_CHILD_RID))";

    private static final String dependencyVersionIDSequence =
            "CREATE SEQUENCE REG_DEP_VERSION_SEQ START WITH 1 INCREMENT BY 1 NOCACHE";

    private static final String dependencyVersionIDTrigger =
            "CREATE OR REPLACE TRIGGER REG_DEPENDENCY_VERSION_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON REG_DEPENDENCY_VERSION " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.REG_PARENT_RID IS NOT NULL AND :NEW.REG_PARENT_VERSION IS NOT NULL " +
                    "AND :NEW.REG_CHILD_RID IS NOT NULL THEN " +
                    "SELECT REG_DEP_VERSION_SEQ.nextval INTO :NEW.REG_DEPENDENCY_VERSION_ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    private static final String contentVersionTable =
            "CREATE TABLE REG_CONTENT_VERSION (" +
                    "REG_CONTENT_VERSION_ID VARCHAR2 (50)," +
                    "REG_CONTENT_DATA BLOB," +
                    "PRIMARY KEY (REG_CONTENT_VERSION_ID))";

    private static final String propertyVersionTable =
            "CREATE TABLE REG_PROPERTY_VERSION (" +
                    "REG_PROPERTY_VERSION_ID INTEGER," +
                    "REG_RID VARCHAR2 (50) NOT NULL," +
                    "REG_VERSION INTEGER NOT NULL," +
                    "REG_NAME VARCHAR2 (100) NOT NULL," +
                    "REG_PROPERTY_VALUE VARCHAR2 (500)," +
                    "PRIMARY KEY (REG_PROPERTY_VERSION_ID)," +
                    "FOREIGN KEY (REG_RID, REG_VERSION) REFERENCES REG_RESOURCE_VERSION (REG_RID, REG_VERSION))";

    private static final String propertyVerionIDSequence =
            "CREATE SEQUENCE REG_PROPERTY_VERSION_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    private static final String propertyVersionIDTrigger =
            "CREATE OR REPLACE TRIGGER REG_PROPERTY_VERSION_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON REG_PROPERTY_VERSION " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.REG_RID IS NOT NULL AND :NEW.REG_VERSION IS NOT NULL AND :NEW.REG_NAME IS NOT NULL THEN " +
                    "SELECT REG_PROPERTY_VERSION_SEQUENCE.nextval INTO :NEW.REG_PROPERTY_VERSION_ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    //////////////////////////////////////////////////////////////////////////////////
    // Tables related to snapshot management                                        //
    //////////////////////////////////////////////////////////////////////////////////

    // UUID is used for snapshot ID, so that snapshots can be achieved and reloaded without
    // colliding with existing snapshots.

    private static final String snapshotTable =
            "CREATE TABLE REG_SNAPSHOT (" +
                    "REG_SNAPSHOT_ID INTEGER," +
                    "REG_ROOT_ID VARCHAR2 (50) NOT NULL," +
                    "PRIMARY KEY (REG_SNAPSHOT_ID)," +
                    "UNIQUE (REG_SNAPSHOT_ID, REG_ROOT_ID))";

    private static final String snapshotIDSequence =
            "CREATE SEQUENCE REG_SNAPSHOT_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    private static final String snapshotIDTrigger =
            "CREATE OR REPLACE TRIGGER REG_SNAPSHOT_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON REG_SNAPSHOT " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.REG_ROOT_ID IS NOT NULL THEN " +
                    "SELECT REG_SNAPSHOT_SEQUENCE.nextval INTO :NEW.REG_SNAPSHOT_ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    private static final String snapshotRootIDIndex =
            "CREATE INDEX REG_INDEX_SNAPSHOT_ROOT_ID ON REG_SNAPSHOT (REG_ROOT_ID)";

    private static final String snapshotResourceVersionTable =
            "CREATE TABLE REG_SNAPSHOT_RESOURCE_VERSION (" +
                    "REG_SRV_ID INTEGER," +
                    "REG_SNAPSHOT_ID INTEGER NOT NULL," +
                    "REG_RID VARCHAR2 (50) NOT NULL," +
                    "REG_VERSION INTEGER NOT NULL," +
                    "PRIMARY KEY (REG_SRV_ID)," +
                    "UNIQUE (REG_SNAPSHOT_ID, REG_RID, REG_VERSION))";

    private static final String srvIDSequence =
            "CREATE SEQUENCE REG_SNAP_RESOURCE_VER_SEQ START WITH 1 INCREMENT BY 1 NOCACHE";

    private static final String srvIDTrigger =
            "CREATE OR REPLACE TRIGGER REG_SNAP_RESOURCE_VER_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON REG_SNAPSHOT_RESOURCE_VERSION " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.REG_SNAPSHOT_ID IS NOT NULL AND :NEW.REG_RID IS NOT NULL AND :NEW.REG_VERSION IS NOT NULL THEN " +
                    "SELECT REG_SNAP_RESOURCE_VER_SEQ.nextval INTO :NEW.REG_SRV_ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    ////////////////////////////////////////////////////////////////////////////////////
    // User manager tables                                                            //
    ////////////////////////////////////////////////////////////////////////////////////

    public static String userTable = "CREATE TABLE UM_USERS ("
            + "ID INTEGER," + "USER_NAME VARCHAR2(255) NOT NULL,"
            + "USER_PASSWORD VARCHAR2(255) NOT NULL, " + "PRIMARY KEY (ID), " + "UNIQUE(USER_NAME))";

    public static final String userIDSequence =
            "CREATE SEQUENCE UM_USERS_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    public static final String userIDTrigger =
            "CREATE OR REPLACE TRIGGER UM_USERS_TRIGGER " +
		            "BEFORE INSERT " +
                    "ON UM_USERS " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.USER_NAME IS NOT NULL AND :NEW.USER_PASSWORD IS NOT NULL THEN " +
                    "SELECT UM_USERS_SEQUENCE.nextval INTO :NEW.ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    public static String userAttributeTable = "CREATE TABLE UM_USER_ATTRIBUTES ("
            + "ID INTEGER," + "ATTR_NAME VARCHAR2(255) NOT NULL,"
            + "ATTR_VALUE VARCHAR2(255), " + "USER_ID INTEGER, "
            + "FOREIGN KEY (USER_ID) REFERENCES UM_USERS(ID) ON DELETE CASCADE,"
            + "PRIMARY KEY (ID))";

    public static final String userAttributeIDSeq =
            "CREATE SEQUENCE UM_USER_ATTRIBUTES_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    public static final String userAttributeTrigger =
            "CREATE OR REPLACE TRIGGER UM_USER_ATTRIBUTES_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON UM_USER_ATTRIBUTES " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.ATTR_NAME IS NOT NULL THEN " +
                    "SELECT UM_USER_ATTRIBUTES_SEQUENCE.nextval INTO :NEW.ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    public static String roleTable = "CREATE TABLE UM_ROLES ("
            + "ID INTEGER, " + "ROLE_NAME VARCHAR2(255) NOT NULL, "
            + "PRIMARY KEY (ID), " + "UNIQUE(ROLE_NAME))";

    public static final String roleIDSeq =
            "CREATE SEQUENCE UM_ROLES_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    public static final String roleIDTrigger =
            "CREATE OR REPLACE TRIGGER UM_ROLES_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON UM_ROLES " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.ROLE_NAME IS NOT NULL THEN " +
                    "SELECT UM_ROLES_SEQUENCE.nextval INTO :NEW.ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    public static String roleAttributeTable = "CREATE TABLE UM_ROLE_ATTRIBUTES ("
            + "ID INTEGER, " + "ATTR_NAME VARCHAR2(255) NOT NULL, "
            + "ATTR_VALUE VARCHAR2(255), " + "ROLE_ID INTEGER, "
            + "FOREIGN KEY (ROLE_ID) REFERENCES UM_ROLES(ID) ON DELETE CASCADE,"
            + "PRIMARY KEY (ID))";

    public static final String roleAttributesIDSeq =
            "CREATE SEQUENCE UM_ROLE_ATTRIBUTES_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    public static final String roleAttributesIDTrigger =
            "CREATE OR REPLACE TRIGGER UM_ROLE_ATTRIBUTES_TRIGGER " +
		            "BEFORE INSERT " +
                    "ON UM_ROLE_ATTRIBUTES " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.ATTR_NAME IS NOT NULL THEN " +
                    "SELECT UM_ROLE_ATTRIBUTES_SEQUENCE.nextval INTO :NEW.ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    public static String permissionTable = "CREATE TABLE UM_PERMISSIONS ("
            + "ID INTEGER, "
            + "RESOURCE_ID VARCHAR2(255) NOT NULL, " + "ACTION VARCHAR2(255) NOT NULL, "
            + "PRIMARY KEY (ID))";

    public static final String permissionIDSeq =
            "CREATE SEQUENCE UM_PERMISSIONS_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    public static final String permissionIDTrigger =
            "CREATE OR REPLACE TRIGGER UM_PERMISSIONS_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON UM_PERMISSIONS " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.RESOURCE_ID IS NOT NULL AND :NEW.ACTION IS NOT NULL THEN " +
                    "SELECT UM_PERMISSIONS_SEQUENCE.nextval INTO :NEW.ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    public static String rolePermissionTable = "CREATE TABLE UM_ROLE_PERMISSIONS ( "
            + "ID INTEGER, " + "PERMISSION_ID INTEGER NOT NULL, "
            + "ROLE_ID INTEGER NOT NULL, " + "IS_ALLOWED SMALLINT NOT NULL, "
            + "UNIQUE (PERMISSION_ID, ROLE_ID), "
            + "FOREIGN KEY (PERMISSION_ID) REFERENCES UM_PERMISSIONS(ID) ON DELETE  CASCADE, "
            + "FOREIGN KEY (ROLE_ID) REFERENCES UM_ROLES(ID) ON DELETE CASCADE, "
            + "PRIMARY KEY (ID))";

    public static final String rolePermissionIDSeq =
            "CREATE SEQUENCE UM_ROLE_PERMISSIONS_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    public static final String rolePermissionIDTrigger =
            "CREATE OR REPLACE TRIGGER UM_ROLE_PERMISSIONS_TRIGGER " +
		            "BEFORE INSERT " +
                    "ON UM_ROLE_PERMISSIONS " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.PERMISSION_ID IS NOT NULL AND :NEW.ROLE_ID IS NOT NULL " +
                    "AND :NEW.IS_ALLOWED IS NOT NULL THEN " +
                    "SELECT UM_ROLE_PERMISSIONS_SEQUENCE.nextval INTO :NEW.ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    public static String userPermissionTable = "CREATE TABLE UM_USER_PERMISSIONS ("
            + "ID INTEGER, " + "PERMISSION_ID INTEGER NOT NULL, "
            + "USER_ID INTEGER NOT NULL, " + "IS_ALLOWED SMALLINT NOT NULL, "
            + "UNIQUE (PERMISSION_ID, USER_ID), "
            + "FOREIGN KEY (PERMISSION_ID) REFERENCES UM_PERMISSIONS(ID) ON DELETE CASCADE, "
            + "FOREIGN KEY (USER_ID) REFERENCES UM_USERS(ID) ON DELETE CASCADE, "
            + "PRIMARY KEY (ID))";

    public static final String userPermissionIDSeq =
            "CREATE SEQUENCE UM_USER_PERMISSIONS_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    public static final String userPermissionIDTrigger =
            "CREATE OR REPLACE TRIGGER UM_USER_PERMISSIONS_TRIGGER " +
		            "BEFORE INSERT " +
		            "ON UM_USER_PERMISSIONS " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.PERMISSION_ID IS NOT NULL AND :NEW.USER_ID IS NOT NULL " +
                    "AND :NEW.IS_ALLOWED IS NOT NULL THEN " +
                    "SELECT UM_USER_PERMISSIONS_SEQUENCE.nextval INTO :NEW.ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    public static String userRolesTable = "CREATE TABLE UM_USER_ROLES ("
            + "ID INTEGER, " + "ROLE_ID INTEGER NOT NULL, "
            + "USER_ID INTEGER NOT NULL, " + "UNIQUE (USER_ID, ROLE_ID), "
            + "FOREIGN KEY (ROLE_ID) REFERENCES UM_ROLES(ID) ON DELETE CASCADE, "
            + "FOREIGN KEY (USER_ID) REFERENCES UM_USERS(ID) ON DELETE CASCADE, "
            + "PRIMARY KEY (ID))";

    public static final String userRolesIDSeq =
            "CREATE SEQUENCE UM_USER_ROLES_SEQUENCE START WITH 1 INCREMENT BY 1 NOCACHE";

    public static final String userRolesIDTrigger =
            "CREATE OR REPLACE TRIGGER UM_USER_ROLES_TRIGGER " +
                    "BEFORE INSERT " +
                    "ON UM_USER_ROLES " +
                    "REFERENCING NEW AS NEW " +
                    "FOR EACH ROW " +
                    "BEGIN " +
                    "IF :NEW.ROLE_ID IS NOT NULL AND :NEW.USER_ID IS NOT NULL THEN " +
                    "SELECT UM_USER_ROLES_SEQUENCE.nextval INTO :NEW.ID FROM dual; " +
                    "END IF; " +
                    "END; ";

    public OracleDatabaseCreator(DataSource dataSource) {
        super(dataSource);
    }

    public String getClusterLockTable() {
        return clusterLockTable;
    }

    public String getResourceTable() {
        return resourceTable;
    }

    public String getContentTable() {
        return contentTable;
    }

    public String getDependencyTable() {
        return dependencyTable;
    }

    public String getPropertyTable() {
        return propertyTable;
    }

    public String getAssociationTable() {
        return associationTable;
    }

    public String getTagTable() {
        return tagTable;
    }

    public String getCommentTable() {
        return commentTable;
    }

    public String getRatingTable() {
        return ratingTable;
    }

    public String getLogTable() {
        return logTable;
    }

    public String getResourceVersionTable() {
        return resourceVersionTable;
    }

    public String getDependencyVersionTable() {
        return dependencyVersionTable;
    }

    public String getContentVersionTable() {
        return contentVersionTable;
    }

    public String getPropertyVersionTable() {
        return propertyVersionTable;
    }

    public String getSnapshotTable() {
        return snapshotTable;
    }

    public String getSnapshotResourceVersionTable() {
        return snapshotResourceVersionTable;
    }

    public String getUserTable() {
        return userTable;
    }

    public String getUserAttributeTable() {
        return userAttributeTable;
    }

    public String getRoleTable() {
        return roleTable;
    }

    public String getRoleAttributeTable() {
        return roleAttributeTable;
    }

    public String getPermissionTable() {
        return permissionTable;
    }

    public String getRolePermissionTable() {
        return rolePermissionTable;
    }

    public String getUserPermissionTable() {
        return userPermissionTable;
    }

    public String getUserRolesTable() {
        return userRolesTable;
    }

    public String getSnapshotRootIDIndex() {
        return snapshotRootIDIndex;
    }

    public void executeResourceStoreDatabaseSpecificStatements(Statement s) throws SQLException {

        s.executeUpdate(dependencySequence);
        s.executeUpdate(dependencyIDIncrementTrigger);
        s.executeUpdate(propertySequence);
        s.executeUpdate(propertyIDTrigger);
        s.executeUpdate(associationSequence);
        s.executeUpdate(associationIDTrigger);
        s.executeUpdate(tagIDSequence);
        s.executeUpdate(tagIDTrigger);
        s.executeUpdate(commentIDSequence);
        s.executeUpdate(commentIDTrigger);
        s.executeUpdate(ratingIDSequence);
        s.executeUpdate(ratingIDTrigger);
        s.executeUpdate(logIDSequence);
        s.executeUpdate(logIDTrigger);
        s.executeUpdate(resourceVersionIDSequence);
        s.executeUpdate(resourceVersionIDTrigger);
        s.executeUpdate(dependencyVersionIDSequence);
        s.executeUpdate(dependencyVersionIDTrigger);
        s.executeUpdate(propertyVerionIDSequence);
        s.executeUpdate(propertyVersionIDTrigger);
        s.executeUpdate(snapshotIDSequence);
        s.executeUpdate(snapshotIDTrigger);
        s.executeUpdate(srvIDSequence);
        s.executeUpdate(srvIDTrigger);

        s.executeUpdate(resourcePathIndex);
        s.executeUpdate(sourcePathIndex);
        s.executeUpdate(targetPathIndex);
    }

    public void executeUserStoreDatabaseSpecificStatements(Statement s) throws SQLException {

        s.executeUpdate(userIDSequence);
        s.executeUpdate(userIDTrigger);
        s.executeUpdate(userAttributeIDSeq);
        s.executeUpdate(userAttributeTrigger);
        s.executeUpdate(roleIDSeq);
        s.executeUpdate(roleIDTrigger);
        s.executeUpdate(roleAttributesIDSeq);
        s.executeUpdate(roleAttributesIDTrigger);
        s.executeUpdate(permissionIDSeq);
        s.executeUpdate(permissionIDTrigger);
        s.executeUpdate(userPermissionIDSeq);
        s.executeUpdate(userPermissionIDTrigger);
        s.executeUpdate(rolePermissionIDSeq);
        s.executeUpdate(rolePermissionIDTrigger);
        s.executeUpdate(userRolesIDSeq);
        s.executeUpdate(userRolesIDTrigger);
    }
}
