/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.server;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.utils.AntBuildInvoker;
import org.wso2.carbon.CarbonConstants;

import java.io.File;

/**
 * The main class of the standalone server runtime.
 */
public class Main {
    private static final Log log = LogFactory.getLog(Main.class);

    // Server commands
    public static final String COMMAND_RUN = "RUN";
    public static final String COMMAND_DEBUG = "DEBUG";

    public static Thread shutdownHookThread;
    private TomcatServer tomcatServer;

    private static final String CARBON_HOME = "carbon.home";
    private static final String AXIS2_HOME = "axis2.home";

    public Main() {
        if (System.getProperty(CARBON_HOME) == null) {
            System.setProperty(CARBON_HOME, ".");
        }
        System.setProperty(AXIS2_HOME, System.getProperty(CARBON_HOME));
    }


    public void startServer() {
        try {
            tomcatServer = new TomcatServer();
            tomcatServer.start();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public void shutdown() throws Exception {
        tomcatServer.stop();
    }

    public static void main(String[] args) {
        System.setProperty(CarbonConstants.START_TIME, System.currentTimeMillis() + "");
        String cmd = null;

        // Set the System properties
        for (String arg : args) {
            if (arg.startsWith("-D")) {
                int indexOfEq = arg.indexOf("=");
                String property;
                String value;
                if (indexOfEq != -1) {
                    property = arg.substring(2, indexOfEq);
                    value = arg.substring(indexOfEq + 1);
                } else {
                    property = arg.substring(2);
                    value = "true";
                }
                System.setProperty(property, value);
            } else {
                if (cmd == null) {
                    cmd = arg;
                }
            }
        }

        if (cmd == null) {
            cmd = COMMAND_RUN;
        }

        final Main main = new Main();

        if (cmd.toUpperCase().endsWith(COMMAND_RUN) ||
            cmd.toUpperCase().endsWith(COMMAND_DEBUG)) {
            log.info("Starting WSO2 Carbon...");
            log.info("Using Java Home        : " + System.getProperty("java.home"));
            log.info("Using Java Version     : " + System.getProperty("java.version"));
            String carbonHome;
            if ((carbonHome = System.getProperty("carbon.home")).equals(".")) {
                carbonHome = new File(".").getAbsolutePath();
            }
            log.info("Using Carbon Home      : " + carbonHome);

            // Run setup which will setup the system when running it for the first time
            if (System.getProperty("setup") != null) {
                main.setup();
            }

            main.startServer();
        } else {
            printUsages();
        }
    }

    private void setup() {
        Thread th = new Thread() {
            @Override
            public void run() {
                log.info("Setting up the system...");
                try {
                    File buildFile = new File(System.getProperty(CARBON_HOME) +
                                              File.separator + "bin" + File.separator + "build.xml");
                    new AntBuildInvoker(buildFile).invokeDefaultTarget(false);
                } catch (Exception e) {
                    log.info("Skipping setup. Reason " + e.getMessage());
                }
            }
        };
        th.setPriority(Thread.MIN_PRIORITY);
        th.start();
    }

    private static void printUsages() {
        String osName = System.getProperty("os.name");

        System.out.println("Usage: wso2server." + ((osName.toLowerCase().indexOf("win") == -1) ? "sh" : "bat ") +
                           " [command]");
        System.out.println("command:");
        System.out.println("\t--debug <port> \tStart the server in remote debugging mode." +
                           "\n\t\t\tport: The remote debugging port.");
        System.out.println("\t--run\t\tRun the server");
        if (osName.toLowerCase().indexOf("win") == -1) {
            System.out.println("\t--start\t\tStart Carbon as a Unix daemon in the background");
            System.out.println("\t--console\tStart Carbon as a Unix daemon in the foreground");
            System.out.println("\t--stop\t\tStop the Carbon Unix daemon");
            System.out.println("\t--status\tGet the status of the Carbon Unix daemon");
            System.out.println("\t--dump\t\tPrint a thread dump of the Carbon Unix daemon");
            System.out.println("\t--restart\tRestart the Carbon Unix daemon");
        }
        System.out.println("\t--version\tWhat version of the Carbon server are you running");
        System.out.println();
    }
}
