/*                                                                             
 * Copyright 2004,2005 The Apache Software Foundation.                         
 *                                                                             
 * Licensed under the Apache License, Version 2.0 (the "License");             
 * you may not use this file except in compliance with the License.            
 * You may obtain a copy of the License at                                     
 *                                                                             
 *      http://www.apache.org/licenses/LICENSE-2.0                             
 *                                                                             
 * Unless required by applicable law or agreed to in writing, software         
 * distributed under the License is distributed on an "AS IS" BASIS,           
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    
 * See the License for the specific language governing permissions and         
 * limitations under the License.                                              
 */
package org.wso2.carbon.server.transports;

import org.w3c.dom.*;
import org.wso2.carbon.utils.ServerConstants;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import java.io.File;
import java.util.*;
import java.net.SocketException;
import java.net.NetworkInterface;
import java.net.InetAddress;

/**
 * This class is reponsible for starting up the transports when Carbon is running
 * in standalone mode
 */
public class TransportManager {

    private static Map<String, Transport> transports = new HashMap<String, Transport>();

    private static String hostAddress = "127.0.0.1";

    /**
     * Initialize the TransportManager. This will load all the transports and start them up.
     */
    public static void init() {
        String transportsFile = getTransportsFilePath();

        try {
            File file = new File(transportsFile);
            if(!file.exists()){
                return;
            }
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document doc = db.parse(file);
            doc.getDocumentElement().normalize();
            Element rootEle = doc.getDocumentElement();
            Attr hostNameAttr = rootEle.getAttributeNode("hostName");
            if (hostNameAttr != null) {
                String hostNameAttrValue = hostNameAttr.getValue();
                if (hostNameAttrValue != null && hostNameAttrValue.length() != 0) {
                    hostAddress = hostNameAttrValue;
                }
            }
            NodeList childNodes = rootEle.getElementsByTagName("transport");
            for (int i = 0; i < childNodes.getLength(); i++) {
                Element transportEle = (Element) childNodes.item(i);
                NamedNodeMap transportAttribs = transportEle.getAttributes();
                Node nameAttrib = transportAttribs.getNamedItem("name");
                String transport;
                if (nameAttrib == null ||
                    (transport = nameAttrib.getNodeValue()) == null ||
                    transport.trim().length() == 0) {
                    throw new RuntimeException("Required attribute, transport name is not specified.");
                }

                Node classAttrib = transportAttribs.getNamedItem("class");
                String clazz;
                if (classAttrib == null ||
                    (clazz = classAttrib.getNodeValue()) == null ||
                    clazz.trim().length() == 0) {
                    throw new RuntimeException("Required attribute, transport class is not specified.");
                }
                Transport transportImpl = (Transport) Class.forName(clazz).newInstance();
                transports.put(transport, transportImpl);

                List<TransportParameter> transportParameters = new ArrayList<TransportParameter>();
                NodeList parameterList = transportEle.getElementsByTagName("parameter");
                for (int j = 0; j < parameterList.getLength(); j++) {
                    Element paramEle = (Element) parameterList.item(j);
                    String value = null;
                    NodeList values = paramEle.getChildNodes();
                    for (int k = 0; k < values.getLength(); k++) {
                        Node node = values.item(k);
                        if ((value = node.getNodeValue().trim()).length() != 0) {
                            break;
                        }
                    }
                    value = replaceSystemProperty(value);
                    String paramName = paramEle.getAttribute("name");
                    transportParameters.add(new TransportParameter(paramName, value));
                    if(paramName.equals("port")){
                        System.setProperty("carbon." + transport + ".port", value);
                    }
                }
                transportImpl.init(transportParameters);
//                transportImpl.start();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private static String replaceSystemProperty(String text) {
        int indexOfStartingChars;
        int indexOfClosingBrace;

        // The following condition deals with properties.
        // Properties are specified as ${system.property},
        // and are assumed to be System properties
        if ((indexOfStartingChars = text.indexOf("${")) != -1 &&
            (indexOfClosingBrace = text.indexOf("}")) != -1) { // Is a property used?
            String sysProp = text.substring(indexOfStartingChars + 2,
                                            indexOfClosingBrace);
            String propValue = System.getProperty(sysProp);
            if (propValue != null) {
                text = text.substring(0, indexOfStartingChars) + propValue +
                       text.substring(indexOfClosingBrace + 1);
            }
            if(sysProp.equals(ServerConstants.CARBON_HOME)){
                if(System.getProperty(ServerConstants.CARBON_HOME).equals(".")){
                    text = new File(".").getAbsolutePath() + File.separator + text;
                }
            }
        }
        return text;
    }
    
    private static String getTransportsFilePath(){
    	String transportsPath= System.getProperty(ServerConstants.CARBON_HOME) + File.separator +
        	"conf" + File.separator + "transports.xml";
    	try{
    		String tranportsProperty = System.getProperty(ServerConstants.TRANSPORTS_XML_PATH);
    		new File(tranportsProperty);
    		//should check whether the file present, but should accomodate 
    		//for the special case of this running as a war
    		transportsPath=tranportsProperty;
    	}catch(Exception e){
    		//the property is invalid does not exist keep using the default file path
    	}
    	return transportsPath;
    }
    
    /**
     * Get the transport port
     *
     * @param transport The transport
     * @return The port corresponding to the transport. If the transport is not available, or there
     *         is no port for the transport, returns -1
     */
    public int getPort(String transport) {

        Transport transportImpl = transports.get(transport);
        if (transportImpl != null) {
            return transportImpl.getPort();
        }
        return -1;
    }

    /**
     * Get the transport proxy port
     *
     * @param transport The transport
     * @return The proxy port corresponding to the transport. If the transport is not available, or there
     *         is no proxy port for the transport, returns -1
     */
    public int getProxyPort(String transport) {

        Transport transportImpl = transports.get(transport);
        if (transportImpl != null) {
            return transportImpl.getProxyPort();
        }
        return -1;
    }

    public void startTransport(String transport) throws Exception {
        Transport transportImpl = transports.get(transport);
        if (transportImpl != null) {
            transportImpl.start();
        } else {
            if (!System.getProperty(ServerConstants.REPO_WRITE_MODE, "false").equals("true")) {
                throw new RuntimeException("Transport " + transport + " not found");
            }
        }
    }

    public void stopTransport(String transport) throws Exception {
        Transport transportImpl = transports.get(transport);
        if (transportImpl != null) {
            transportImpl.stop();
        } else {
            if (!System.getProperty(ServerConstants.REPO_WRITE_MODE, "false").equals("true")) {
                throw new RuntimeException("Transport " + transport + " not found");
            }
        }
    }

    /**
         * Returns the ip address to be used for the replyto epr
         * CAUTION:
         * This will go through all the available network interfaces and will try to return an ip address.
         * First this will try to get the first IP which is not loopback address (127.0.0.1). If none is found
         * then this will return this will return 127.0.0.1.
         * This will <b>not<b> consider IPv6 addresses.
         * <p/>
         *
         * @return Returns String.
         * @throws java.net.SocketException   SocketException
         */
    public static String getIpAddress() throws SocketException {
        if (!hostAddress.equals("127.0.0.1")) {
            Enumeration e = NetworkInterface.getNetworkInterfaces();
            while (e.hasMoreElements()) {
                NetworkInterface netface = (NetworkInterface) e.nextElement();
                Enumeration addresses = netface.getInetAddresses();

                while (addresses.hasMoreElements()) {
                    InetAddress ip = (InetAddress) addresses.nextElement();
                    if (!ip.isLoopbackAddress()) {
                        return ip.getHostAddress();
                    }
                }
            }
        }
        return hostAddress;
    }

}
