/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.user.core.def;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.user.core.Authorizer;
import org.wso2.carbon.user.core.UserCoreConstants;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.def.util.DefaultCommons;
import org.wso2.carbon.user.core.def.util.DefaultSQLHelperImpl;
import org.wso2.carbon.user.core.i18n.Messages;

/**
 * @see org.wso2.usermanager.Authorizer
 */
public class DefaultAuthorizer implements Authorizer {

    /** DBCP datasource */
    protected DataSource dataSource = null;

    /** logger */
    private static Log log = LogFactory.getLog(DefaultAuthorizer.class);

    /** Algorithm to authorize users */
    private String algorithm = null;

    protected SQLHelper sqlHelper = new DefaultSQLHelperImpl();

    protected CacheManager cacheManager = null;

    private DefaultAuthorizer() {
	cacheManager = new CacheManager();
	Cache cache1 = new Cache(UserCoreConstants.AUTHZ_CACHE,
		UserCoreConstants.MAX_OBJECTS_IN_CACHE, false, false, UserCoreConstants.TTL_CACHE,
		UserCoreConstants.IDLE_TIME_IN_CACHE);
	cacheManager.addCache(cache1);
    }

    public DefaultAuthorizer(DataSource dataSource, String algo) {
	this();
	this.dataSource = dataSource;
	this.algorithm = algo;
    }

    public DefaultAuthorizer(DataSource dataSource, String algo, SQLHelper sqlHelper) {
	this();
	this.dataSource = dataSource;
	this.algorithm = algo;
	this.sqlHelper = sqlHelper;
    }

    public boolean isRoleAuthorized(String roleName, String resourceId, String action)
	    throws UserStoreException {

	Boolean isAuthorized = getRoleAuthorized(roleName, resourceId, action, null);
	return isAuthorized != null && isAuthorized;
    }

    public boolean isUserAuthorized(String userName, String resourceId, String action)
	    throws UserStoreException {

	Boolean isAuthorized = null;

	Cache cache = cacheManager.getCache(UserCoreConstants.AUTHZ_CACHE);
	Element element = cache.get(userName + resourceId + action);

	if (element != null) {
	    Boolean bool = (Boolean) element.getObjectValue();
	    return bool.booleanValue();
	}

	if (algorithm.equals(DefaultRealmConfig.PERMISSION_USER_ONLY)) {
	    isAuthorized = getUserAuthorized(userName, resourceId, action, null);
	} else if (algorithm.equals(DefaultRealmConfig.PERMISSION_BLOCK_FIRST)) {
	    isAuthorized = getUserAuthorizationConsideringRoles(userName, resourceId, action);
	}

	boolean value = false;
	if (isAuthorized != null && isAuthorized) {
	    value = true;
	}

	element = new Element(userName + resourceId + action, value);
	cache.put(element);

	return value;

    }

    public String[] getAllowedUsersForResource(String resourceId, String action)
	    throws UserStoreException {
	return getUsersOnResource(resourceId, action, 1);
    }

    public String[] getDeniedUsersForResource(String resourceId, String action)
	    throws UserStoreException {
	return getUsersOnResource(resourceId, action, 0);
    }

    public String[] getAllowedRolesForResource(String resourceId, String action)
	    throws UserStoreException {
	return getAllowedRolesForResource(resourceId, action, 1);
    }

    public String[] getDeniedRolesForResource(String resourceId, String action)
	    throws UserStoreException {
	return getAllowedRolesForResource(resourceId, action, 0);
    }

    protected Boolean getUserAuthorizationConsideringRoles(String userName, String resourceId,
	    String action) throws UserStoreException {
	Connection dbConnection = null;
	Boolean isAuthorized = false;
	boolean onceAllowed = false;
	PreparedStatement stmt = null;
	try {
	    dbConnection = dataSource.getConnection();
	    dbConnection.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    isAuthorized = getUserAuthorized(userName, resourceId, action, dbConnection);
	    if (isAuthorized != null)
		return isAuthorized;

	    String roles[] = new String[0];
	    DefaultCommons commons = new DefaultCommons();
	    roles = commons.getUserRoles(sqlHelper, dbConnection, userName);

	    stmt = dbConnection.prepareStatement(sqlHelper
		    .getSQL(DefaultRealmConstants.GET_ROLES_PERMISSION_FOR_RESOURCE));
	    stmt.setString(1, resourceId);
	    stmt.setString(2, action);
	    ResultSet rs = stmt.executeQuery();

	    Map<String, Boolean> map = new HashMap<String, Boolean>();
	    while (rs.next()) {
		int allowed = rs.getShort(sqlHelper
			.getColumnName(DefaultRealmConstants.COLUMN_ID_IS_ALLOWED));
		String role = rs.getString(sqlHelper
			.getColumnName(DefaultRealmConstants.COLUMN_ID_ROLE_NAME));
		boolean value = false;
		if (allowed == 1) {
		    value = true;
		}
		map.put(role, value);
	    }

	    for (String role : roles) {
		Boolean value = map.get(role);
		if (value != null) {
		    if (value.equals(false)) {
			isAuthorized = false;
			break;
		    } else {
			isAuthorized = true;
		    }
		}
	    }

	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if(stmt != null){
		    stmt.close();
		}
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}

	return isAuthorized;
    }

    public Boolean getRoleAuthorized(String roleName, String resourceId, String action,
	    Connection conn) throws UserStoreException {
	Boolean isAuthorized = null;
	Connection dbConnection = null;
	boolean dbConnectionReceived = false;
	try {
	    if (conn != null) {
		dbConnectionReceived = true;
		dbConnection = conn;
	    } else {
		dbConnection = dataSource.getConnection();
		if (dbConnection == null) {
		    throw new UserStoreException(Messages.getMessage("null_connection"));
		}
	    }
	    dbConnection.setAutoCommit(false);
	    PreparedStatement isRoleAuthorizedStmt = dbConnection.prepareStatement(sqlHelper
		    .getSQL(DefaultRealmConstants.GET_ROLE_AUTHORIZED));
	    isRoleAuthorizedStmt.setString(1, resourceId);
	    isRoleAuthorizedStmt.setString(2, action);
	    isRoleAuthorizedStmt.setString(3, roleName);
	    ResultSet rs = isRoleAuthorizedStmt.executeQuery();
	    if (rs.next()) {
		isAuthorized = rs.getBoolean(DefaultRealmConstants.COLUMN_NAME_IS_ALLOWED);
	    }
	    isRoleAuthorizedStmt.close();
	} catch (SQLException e) {
	    log.debug(e);
	    e.printStackTrace();
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if ((!dbConnectionReceived) && dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return isAuthorized;
    }

    public Boolean getUserAuthorized(String userName, String resourceId, String action,
	    Connection conn) throws UserStoreException {
	Boolean result = null;
	Connection dbConnection = null;
	boolean dbConnectionReceived = false;
	try {
	    if (conn != null) {
		dbConnection = conn;
		dbConnectionReceived = true;
	    } else {
		dbConnection = dataSource.getConnection();
	    }

	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    PreparedStatement isUserAuthorizedStmt = dbConnection.prepareStatement(sqlHelper
		    .getSQL(DefaultRealmConstants.GET_USER_AUTHORIZED));
	    isUserAuthorizedStmt.setString(1, resourceId);
	    isUserAuthorizedStmt.setString(2, action);
	    isUserAuthorizedStmt.setString(3, userName);
	    ResultSet rs = isUserAuthorizedStmt.executeQuery();
	    if (rs.next()) {
		result = rs.getBoolean(DefaultRealmConstants.COLUMN_NAME_IS_ALLOWED);
	    }
	    isUserAuthorizedStmt.close();
	} catch (SQLException e) {
	    log.debug(e);
	    e.printStackTrace();
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if ((!dbConnectionReceived) && dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return result;
    }

    protected String[] getUsersOnResource(String resourceId, String action, int isAllowed)
	    throws UserStoreException {
	String[] names = new String[0];
	Connection dbConnection = null;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    PreparedStatement prepStmt = dbConnection.prepareStatement(sqlHelper
		    .getSQL(DefaultRealmConstants.GET_USERS_ON_RESOURCE));
	    prepStmt.setString(1, resourceId);
	    prepStmt.setString(2, action);
	    prepStmt.setShort(3, (short) isAllowed);
	    ResultSet rs = prepStmt.executeQuery();
	    List lst = new LinkedList();
	    String colName = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_USER_NAME);
	    while (rs.next()) {
		lst.add(rs.getString(colName));
	    }
	    if (lst.size() > 0) {
		names = (String[]) lst.toArray(new String[lst.size()]);
	    }
	    prepStmt.close();
	} catch (SQLException e) {
	    e.printStackTrace();
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return names;

    }

    protected String[] getAllowedRolesForResource(String resourceId, String action, int isAllowed)
	    throws UserStoreException {
	String[] names = new String[0];
	Connection dbConnection = null;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    PreparedStatement getAuthorizedRolesForResourceStmt = dbConnection
		    .prepareStatement(sqlHelper
			    .getSQL(DefaultRealmConstants.GET_ROLES_FOR_RESOURCE));
	    getAuthorizedRolesForResourceStmt.setString(1, resourceId);
	    getAuthorizedRolesForResourceStmt.setString(2, action);
	    getAuthorizedRolesForResourceStmt.setShort(3, (short) isAllowed);
	    ResultSet rs = getAuthorizedRolesForResourceStmt.executeQuery();
	    List lst = new LinkedList();
	    String colName = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_ROLE_NAME);
	    while (rs.next()) {
		lst.add(rs.getString(colName));
	    }
	    if (lst.size() > 0) {
		names = (String[]) lst.toArray(new String[lst.size()]);
	    }
	    getAuthorizedRolesForResourceStmt.close();
	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return names;
    }

    protected Boolean getUserAuthorizationOnlyConsideringRoles(String userName, String resourceId,
	    String action) throws UserStoreException {
	Connection dbConnection = null;
	Boolean isAuthorized = false;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);

	    if (isAuthorized == null) {
		PreparedStatement stmt = dbConnection.prepareStatement(sqlHelper
			.getSQL(DefaultRealmConstants.GET_ROLES_FOR_RESOURCE));
		stmt.setString(1, resourceId);
		stmt.setString(2, action);
		stmt.setString(3, userName);
		ResultSet rs = stmt.executeQuery();
		List lst = new LinkedList();
		String colName = sqlHelper
			.getColumnName(DefaultRealmConstants.COLUMN_ID_IS_ALLOWED);
		short value = 0;
		while (rs.next()) {
		    value = rs.getShort(colName);
		}

		if (value == 1) {
		    isAuthorized = true;
		}

	    }
	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}

	return isAuthorized;
    }

}
