package org.wso2.carbon.user.core.hybrid;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.user.core.Authorizer;
import org.wso2.carbon.user.core.UserCoreConstants;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.UserStoreReader;
import org.wso2.carbon.user.core.def.DefaultAuthorizer;
import org.wso2.carbon.user.core.def.DefaultRealmConfig;
import org.wso2.carbon.user.core.def.DefaultRealmConstants;
import org.wso2.carbon.user.core.def.DefaultUserStoreAdmin;
import org.wso2.carbon.user.core.def.SQLHelper;
import org.wso2.carbon.user.core.i18n.Messages;

public class HybridAuthorizer implements Authorizer {

    /** logger */
    private static Log log = LogFactory.getLog(HybridAuthorizer.class);

    /** Algorithm to authorize users */
    private String algorithm = null;

    private HybridRealmConfig config = null;
    protected SQLHelper sqlHelper = null;
    protected DefaultAuthorizer defAuthorizer = null;
    protected DataSource dataSource = null;

    protected CacheManager cacheManager = null;

    private HybridAuthorizer() {
	cacheManager = new CacheManager();
	Cache cache1 = new Cache(UserCoreConstants.AUTHZ_CACHE,
		UserCoreConstants.MAX_OBJECTS_IN_CACHE, false, false, UserCoreConstants.TTL_CACHE,
		UserCoreConstants.IDLE_TIME_IN_CACHE);
	cacheManager.addCache(cache1);
    }

    public HybridAuthorizer(DataSource dataSource, String algo, HybridRealmConfig config) {
	this();
	this.dataSource = dataSource;
	sqlHelper = new HybridSQLHelper();
	defAuthorizer = new DefaultAuthorizer(dataSource, algo, sqlHelper);
	this.algorithm = algo;
	this.config = config;

    }

    public String[] getAllowedRolesForResource(String resourceId, String action)
	    throws UserStoreException {
	return defAuthorizer.getAllowedRolesForResource(resourceId, action);
    }

    public String[] getAllowedUsersForResource(String resourceId, String action)
	    throws UserStoreException {
	return defAuthorizer.getAllowedUsersForResource(resourceId, action);
    }

    public String[] getDeniedRolesForResource(String resourceId, String action)
	    throws UserStoreException {
	return defAuthorizer.getDeniedRolesForResource(resourceId, action);
    }

    public String[] getDeniedUsersForResource(String resourceId, String action)
	    throws UserStoreException {
	return defAuthorizer.getDeniedUsersForResource(resourceId, action);
    }

    public boolean isRoleAuthorized(String roleName, String resourceId, String action)
	    throws UserStoreException {
	return defAuthorizer.isRoleAuthorized(roleName, resourceId, action);
    }

    public boolean isUserAuthorized(String userName, String resourceId, String action)
	    throws UserStoreException {
	Boolean isAuthorized = null;

	Cache cache = cacheManager.getCache(UserCoreConstants.AUTHZ_CACHE);
	Element element = cache.get(userName + resourceId + action);

	if (element != null) {
	    Boolean bool = (Boolean) element.getObjectValue();
	    return bool.booleanValue();
	}

	if (algorithm.equals(DefaultRealmConfig.PERMISSION_USER_ONLY)) {
	    isAuthorized = defAuthorizer.getUserAuthorized(userName, resourceId, action, null);
	} else if (algorithm.equals(DefaultRealmConfig.PERMISSION_BLOCK_FIRST)) {
	    isAuthorized = getUserAuthorizationConsideringRoles(userName, resourceId, action);
	}

	boolean value = false;
	if (isAuthorized != null && isAuthorized) {
	    value = true;
	}

	element = new Element(userName + resourceId + action, value);
	cache.put(element);

	return value;
    }

    protected Boolean getUserAuthorizationConsideringRoles(String userName, String resourceId,
	    String action) throws UserStoreException {
	Boolean isAuthorized = defAuthorizer.getUserAuthorized(userName, resourceId, action, null);
	Connection dbConnection = null;
	PreparedStatement stmt = null;
	ResultSet rs = null;
	boolean onceAllowed = false;
	if (isAuthorized == null) {
	    String roles[] = new String[0];
	    try {
		UserStoreReader reader = config.getUserStoreReader();
		roles = reader.getUserRoles(userName);

		// get internal roles defined by external roles
		DefaultUserStoreAdmin defUSAdmin = new DefaultUserStoreAdmin(dataSource, sqlHelper);
		String[] intnlRoles = defUSAdmin.getUserRoles(userName);

		int length = intnlRoles.length;
		if (length > 0) {
		    roles = intnlRoles;
		}

		if (length > 0) {
		    List<String> roleList = new ArrayList<String>();

		    for (String role : roles) {
			roleList.add(role);
		    }

		    for (String role : intnlRoles) {
			roleList.add(role);
		    }
		    roles = roleList.toArray(new String[roleList.size()]);
		}

	    } catch (Exception e) {
		log.debug(e);
		throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	    }
	    try {
		dbConnection = dataSource.getConnection();
		stmt = dbConnection.prepareStatement(sqlHelper
			.getSQL(DefaultRealmConstants.GET_ROLES_PERMISSION_FOR_RESOURCE));
		stmt.setString(1, resourceId);
		stmt.setString(2, action);
		rs = stmt.executeQuery();

		Map<String, Boolean> map = new HashMap<String, Boolean>();
		while (rs.next()) {
		    int allowed = rs.getShort(sqlHelper
			    .getColumnName(DefaultRealmConstants.COLUMN_ID_IS_ALLOWED));
		    String role = rs.getString(sqlHelper
			    .getColumnName(DefaultRealmConstants.COLUMN_ID_ROLE_NAME));
		    boolean value = false;
		    if (allowed == 1) {
			value = true;
		    }
		    map.put(role, value);
		}

		for (String role : roles) {
		    Boolean value = map.get(role);
		    if (value != null) {
			if (value.equals(false)) {
			    isAuthorized = false;
			    break;
			} else {
			    isAuthorized = true;
			}
		    }
		}
	    } catch (SQLException e) {
		log.debug(e);
		throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	    } finally {
		try {
		    if(rs != null){
			rs.close();
		    }
		    if(stmt != null){
			stmt.close();
		    }
		    if (dbConnection != null) {
			dbConnection.close();
		    }
		} catch (SQLException e) {
		    throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
		}
	    }

	    /*
	     * for (String roleName : roles) { Boolean roleAuth =
	     * defAuthorizer.getRoleAuthorized(roleName, resourceId, action,
	     * null); if (roleAuth != null) { if (!roleAuth) { isAuthorized =
	     * roleAuth; break; } else { onceAllowed = true; } } }
	     */
	}

	if (isAuthorized == null && onceAllowed) {
	    isAuthorized = true;
	}

	return isAuthorized;
    }

}
