/*
 * Copyright 2004,2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.utils.deployment;

import org.apache.axis2.AxisFault;
import org.apache.axis2.deployment.DeploymentEngine;
import org.apache.axis2.deployment.ModuleBuilder;
import org.apache.axis2.deployment.ModuleDeployer;
import org.apache.axis2.description.AxisModule;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.axis2.util.Utils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMNamespace;
import org.apache.axiom.om.OMElement;
import org.osgi.framework.Bundle;
import org.wso2.carbon.CarbonConstants;

import java.io.IOException;
import java.net.URL;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 *
 */
public class ModuleRegistry {

    private static Log log = LogFactory.getLog(ModuleRegistry.class);
    /**
     *
     */
    private AxisConfiguration axisConfig;

    /**
     *
     */
    private final Lock lock = new ReentrantLock();

    private Map<Bundle, List<AxisModule>> moduleMap;

    public ModuleRegistry(AxisConfiguration axisConfig) {
        this.axisConfig = axisConfig;
        this.moduleMap = new ConcurrentHashMap<Bundle, List<AxisModule>>();
    }


    /**
     * @param bundle Carbon Compoent that needs to be registred
     */
    public void register(Bundle bundle) {
        lock.lock();
        try {

            addModules(bundle);
        } finally {
            lock.unlock();
        }

    }

    private void addModules(Bundle bundle) {

        final String MODULE_DEPLOYER = "moduleDeployer";
        ModuleDeployer deployer =
                (ModuleDeployer) axisConfig.getParameterValue(MODULE_DEPLOYER);
        if (deployer == null) {
            deployer = new ModuleDeployer(axisConfig);
            try {
                axisConfig.addParameter(MODULE_DEPLOYER, deployer);
            } catch (AxisFault e) {
                String msg = "Error while adding the module deployer";
                log.error(msg, e);
                return;
            }
        }
        if (!moduleMap.containsKey(bundle)) {
            Enumeration enumeration = bundle.findEntries("META-INF", "*module.xml", true);
            List<AxisModule> moduleList = null;
            if (enumeration != null) {
                moduleList = new ArrayList<AxisModule>();
            }
            while (enumeration != null && enumeration.hasMoreElements()) {
                try {
                    ClassLoader loader =
                            new BundleClassLoader(bundle, ModuleRegistry.class.getClassLoader());
                    URL url = (URL) enumeration.nextElement();
                    AxisModule axismodule = new AxisModule();

                    axismodule.setModuleClassLoader(loader);
                    ModuleBuilder builder =
                            new ModuleBuilder(url.openStream(), axismodule, axisConfig);
                    Dictionary headers = bundle.getHeaders();
                    String bundleSymbolicName = (String) headers.get("Bundle-SymbolicName");
                    if (bundleSymbolicName != null && bundleSymbolicName.length() != 0) {
                        axismodule.setName(bundleSymbolicName);
                    }
                    String bundleVersion = (String) headers.get("Bundle-Version");
                    if (bundleVersion != null && bundleVersion.length() != 0) {
                        String moduleVersion = "SNAPSHOT";
                        /*
                            Bundle version is defined as
                            version ::=
                                major( '.' minor ( '.' micro ( '.' qualifier )? )? )?
                                major ::= number
                                minor ::= number
                                micro ::= number
                                qualifier ::= ( alphanum | ’_’ | '-' )+

                            Hence, in order to sync up with Axis2 module versioning, which is a floating
                            point number, following logic is used to create the version
                            version := major(.minormircor)
                         */
                        String[] versionSplit = bundleVersion.split("\\.");
                        if (versionSplit.length == 3) {
                            moduleVersion = versionSplit[0] + "." + versionSplit[1] + versionSplit[2];
                        } else if (versionSplit.length == 2) {
                            moduleVersion = versionSplit[0] + "." + versionSplit[1];
                        } else if (versionSplit.length == 1) {
                            moduleVersion = versionSplit[0];
                        }
                        axismodule.setVersion(moduleVersion);
                        //Set a parameter to identiry this module was deployed through a bundle
                        axismodule.addParameter(createManagedModuleParam());
                    }
                    builder.populateModule();
                    axismodule.setParent(axisConfig);
                    AxisModule module = axisConfig.getModule(axismodule.getName());
                    if (module == null) {
                        DeploymentEngine.addNewModule(axismodule, axisConfig);
                        moduleList.add(axismodule);
                        if (log.isDebugEnabled()) {
                            log.debug("Starting any modules in Bundle - " +
                                    bundle.getSymbolicName() + " - Module Name : " +
                                    axismodule.getName() + " - Module Version : " +
                                    axismodule.getVersion());
                        }
                    } else {
                        log.warn("[ModuleRegistry] Module : " + axismodule.getName() +
                                " is already available.");
                    }
                    // set in default map if necessary
                    Utils.calculateDefaultModuleVersion(axisConfig.getModules(), axisConfig);
                } catch (IOException e) {
                    String msg =
                            "Error while building the module from bunlde : " + bundle.getBundleId();
                    log.error(msg, e);
                }

            }
            if (moduleList != null && moduleList.size() > 0) {
                moduleMap.put(bundle, moduleList);
            }
        }

    }

    public void unRegister(Bundle bundle) {
        lock.lock();
        try {
            List<AxisModule> moduleList = moduleMap.get(bundle);
            if (moduleList != null) {
                for (AxisModule axisModule : moduleList) {
                    axisConfig.removeModule(axisModule.getName(), axisModule.getVersion());
                    if (log.isDebugEnabled()) {
                        log.debug("Stopping" + axisModule.getName() + ":" +
                                axisModule.getVersion() + " moduel in Bundle - " +
                                bundle.getSymbolicName());
                    }
                }
                moduleMap.remove(bundle);
            }
        } finally {
            lock.unlock();
        }
    }

    /**
     * Parameter to identiify modules deployed by bundles
     * @return
     */
    private Parameter createManagedModuleParam(){

        Parameter parameter = new Parameter(CarbonConstants.MANAGED_MODULE_PARAM_NAME, Boolean.TRUE.toString());
        OMFactory fac = OMAbstractFactory.getOMFactory();
        OMNamespace ns = fac.createOMNamespace("", "");

        OMElement paramEle = fac.createOMElement("parameter", ns);
        paramEle.addAttribute("name", CarbonConstants.MANAGED_MODULE_PARAM_NAME, ns);
        paramEle.setText(Boolean.TRUE.toString());
        parameter.setParameterElement(paramEle);

        return parameter;
    }

}
